import aiohttp
import requests

from abc import ABC, abstractmethod
from typing import Optional, Dict

class HttpClient(ABC):
    @abstractmethod
    def fetch(self, url: str, options: dict) -> "HttpResponse":
        pass


class HttpResponse:
    def __init__(self, ok: bool, status_code: int, json_data: dict):
        self.ok = ok
        self.status_code = status_code
        self._json_data = json_data

    def json(self):
        return self._json_data


class DefaultHttpClient(HttpClient):
    async def fetch(self, url: str, options: dict) -> HttpResponse:
        async with aiohttp.ClientSession() as session:
            async with session.request(
                method=options["method"],
                url=url,
                headers=options.get("headers", {}),
                json=options.get("data", None),
            ) as response:
                try:
                    json_data = await response.json()
                    return HttpResponse(
                        ok=response.status >= 200 and response.status <= 299,
                        status_code=response.status,
                        json_data={
                            'data': json_data
                        },
                    )
                except Exception as e:
                    return HttpResponse(
                        ok=False,
                        status_code=response.status,
                        json_data={},
                    )

class SyncHttpClient(HttpClient):
    """Synchronous HTTP client compatible with DefaultHttpClient"""

    def __init__(self, default_timeout: int = 30):
        self.default_timeout = default_timeout

    def fetch(self, url: str, options: dict) -> HttpResponse:
        method = options.get("method", "GET")
        headers = options.get("headers", {})
        timeout = options.get("timeout", self.default_timeout)
        data = options.get("data", None)

        try:
            if method.upper() in ["POST", "PUT", "PATCH"] and data is not None:
                response = requests.request(
                    method=method,
                    url=url,
                    headers=headers,
                    json=data,
                    timeout=timeout
                )
            else:
                response = requests.request(
                    method=method,
                    url=url,
                    headers=headers,
                    timeout=timeout
                )

            return self._make_response(response)
        except requests.RequestException as e:
            return self._handle_error(e)

    def _make_response(self, response: requests.Response) -> HttpResponse:
        try:
            json_data = response.json()
            formatted_json = {'data': json_data}
        except (ValueError, requests.exceptions.JSONDecodeError):
            formatted_json = {}

        ok = response.status_code >= 200 and response.status_code <= 299

        return HttpResponse(
            ok=ok,
            status_code=response.status_code,
            json_data=formatted_json
        )

    def _handle_error(self, error: requests.RequestException) -> HttpResponse:
        if isinstance(error, requests.Timeout):
            status_code = 408
        elif isinstance(error, requests.ConnectionError):
            status_code = 503
        elif isinstance(error, requests.HTTPError):
            status_code = error.response.status_code if error.response else 500
        else:
            status_code = 0

        return HttpResponse(
            ok=False,
            status_code=status_code,
            json_data={"error": str(error), "error_type": type(error).__name__}
        )

    def get(self, url: str,
            headers: Optional[Dict[str, str]] = None,
            timeout: Optional[int] = None) -> HttpResponse:
        options = {
            "method": "GET",
            "headers": headers or {},
            "timeout": timeout if timeout is not None else self.default_timeout
        }
        return self.fetch(url, options)

    def post(self, url: str,
             data: Optional[dict] = None,
             headers: Optional[Dict[str, str]] = None,
             timeout: Optional[int] = None) -> HttpResponse:
        options = {
            "method": "POST",
            "headers": headers or {},
            "data": data,
            "timeout": timeout if timeout is not None else self.default_timeout
        }
        return self.fetch(url, options)


def default_sync_http_client() -> SyncHttpClient:
    return SyncHttpClient()


def default_http_client() -> HttpClient:
    return DefaultHttpClient()


from ..fee_model import FeeModel
from abc import ABC
import math
import asyncio

class SatoshisPerKilobyte(FeeModel):
    """
    Represents the "satoshis per kilobyte" transaction fee model.
    """
    
    def __init__(self, value: int):
        """
        Constructs an instance of the sat/kb fee model.
        
        :param value: The number of satoshis per kilobyte to charge as a fee.
        """
        self.value = value

    def compute_fee(self, tx) -> int:
        """
        Computes the fee for a given transaction.
        
        :param tx: The transaction for which a fee is to be computed.
        :returns: The fee in satoshis for the transaction.
        """
        def get_varint_size(i: int) -> int:
            if i > 2 ** 32:
                return 9
            elif i > 2 ** 16:
                return 5
            elif i > 253:
                return 3
            else:
                return 1

        # Compute the (potentially estimated) size of the transaction
        size = 4  # version
        size += get_varint_size(len(tx.inputs))  # number of inputs

        for tx_input in tx.inputs:
            size += 40  # txid, output index, sequence number
            if tx_input.unlocking_script:
                script_length = len(tx_input.unlocking_script.serialize())
            elif tx_input.unlocking_script_template:
                script_length = tx_input.unlocking_script_template.estimated_unlocking_byte_length()
            else:
                raise ValueError('All inputs must have an unlocking script or an unlocking script template for sat/kb fee computation.')
            size += get_varint_size(script_length)  # unlocking script length
            size += script_length  # unlocking script

        size += get_varint_size(len(tx.outputs))  # number of outputs

        for tx_output in tx.outputs:
            size += 8  # satoshis
            length = len(tx_output.locking_script.serialize())
            size += get_varint_size(length)  # script length
            size += length  # script

        size += 4  # lock time

        # We'll use math.ceil to ensure the miners get the extra satoshi.
        fee = math.ceil((size / 1000) * self.value)
        return fee


# from .chain_tracker import ChainTracker
from typing import List, Optional, TypedDict

from .chaintracker import ChainTracker
from .hash import hash256
from .utils import Reader, Writer, to_hex, to_bytes


class MerkleLeaf(TypedDict, total=False):
    offset: int
    hash_str: Optional[str]
    txid: Optional[bool]
    duplicate: Optional[bool]


class MerklePath:
    """
    Represents a Merkle Path, which is used to provide a compact proof of inclusion for a
    transaction in a block. This class encapsulates all the details required for creating
    and verifying Merkle Proofs.

    Attributes:
        block_height (int): The height of the block in which the transaction is included.
        path (List[List[MerkleLeaf]]): A tree structure representing the Merkle Path,
            with each level containing information about the nodes involved in constructing the proof.

    Example:
        # Creating and verifying a Merkle Path
        merkle_path = MerklePath.from_hex('...')
        is_valid = await merkle_path.verify(txid, chain_tracker)

    Description:
        The MerklePath class is useful for verifying transactions in a lightweight and efficient manner without
        needing the entire block data. This class offers functionalities for creating, converting,
        and verifying these proofs.
    """

    def __init__(self, block_height: int, path: List[List[MerkleLeaf]]):
        self.block_height = block_height
        self.path = path

        # store all the legal offsets which we expect given the txid indices.
        legal_offsets = [set() for _ in range(len(self.path))]
        for height, leaves in enumerate(self.path):
            if not leaves and height == 0:
                raise ValueError(f"Empty level at height: {height}")

            offsets_at_this_height = set()
            for leaf in leaves:
                if leaf["offset"] in offsets_at_this_height:
                    raise ValueError(
                        f"Duplicate offset: {leaf['offset']}, at height: {height}"
                    )
                offsets_at_this_height.add(leaf["offset"])

                if height == 0:
                    if not leaf.get("duplicate"):
                        for h in range(1, len(self.path)):
                            legal_offsets[h].add(leaf["offset"] >> h ^ 1)
                else:
                    if leaf["offset"] not in legal_offsets[height]:
                        legal_offsets_at_height = ", ".join(
                            map(str, legal_offsets[height])
                        )
                        raise ValueError(
                            f"Invalid offset: {leaf['offset']}, at height: {height}, "
                            f"with legal offsets: {legal_offsets_at_height}"
                        )

        root = None
        for idx, leaf in enumerate(self.path[0]):
            if idx == 0:
                root = self.compute_root(leaf.get("hash_str"))
            if root != self.compute_root(leaf.get("hash_str")):
                raise ValueError("Mismatched roots")

    @staticmethod
    def from_hex(hex_str: str) -> "MerklePath":
        """
        Creates a MerklePath instance from a hexadecimal string.

        Args:
            hex_str (str): The hexadecimal string representation of the Merkle Path.

        Returns:
            MerklePath: A new MerklePath instance.
        """
        return MerklePath.from_binary(to_bytes(hex_str, "hex"))

    @staticmethod
    def from_reader(reader: Reader) -> "MerklePath":
        """
        Creates a MerklePath instance from a Reader object.

        Args:
            reader (Reader): The Reader object.

        Returns:
            MerklePath: A new MerklePath instance.
        """
        block_height = reader.read_var_int_num()
        tree_height = reader.read_uint8()
        path = [[] for _ in range(tree_height)]

        for level in range(tree_height):
            n_leaves_at_this_height = reader.read_var_int_num()
            while n_leaves_at_this_height:
                offset = reader.read_var_int_num()
                flags = reader.read_uint8()
                leaf = {"offset": offset}

                if flags & 1:
                    leaf["duplicate"] = True
                else:
                    if flags & 2:
                        leaf["txid"] = True
                    leaf["hash_str"] = to_hex(reader.read(32)[::-1])

                path[level].append(leaf)
                n_leaves_at_this_height -= 1

            path[level].sort(key=lambda e: e["offset"])

        return MerklePath(block_height, path)

    @staticmethod
    def from_binary(bump: bytes) -> "MerklePath":
        """
        Creates a MerklePath instance from a bytes object.

        Args:
            bump (bytes): The binary array representation of the Merkle Path.

        Returns:
            MerklePath: A new MerklePath instance.
        """
        reader = Reader(bump)
        return MerklePath.from_reader(reader)

    def to_binary(self) -> bytes:
        """
        Converts the MerklePath to a binary array format.

        Returns:
            bytes: The binary array representation of the Merkle Path.
        """
        writer = Writer()
        writer.write_var_int_num(self.block_height)
        tree_height = len(self.path)
        writer.write_uint8(tree_height)

        for level in range(tree_height):
            n_leaves = len(self.path[level])
            writer.write_var_int_num(n_leaves)

            for leaf in self.path[level]:
                writer.write_var_int_num(leaf["offset"])
                flags = 0
                if leaf.get("duplicate"):
                    flags |= 1
                if leaf.get("txid"):
                    flags |= 2
                writer.write_uint8(flags)

                if not (flags & 1):
                    writer.write(to_bytes(leaf["hash_str"], "hex")[::-1])

        return writer.to_bytes()

    def to_hex(self) -> str:
        """
        Converts the MerklePath to a hexadecimal string format.

        Returns:
            str: The hexadecimal string representation of the Merkle Path.
        """
        return to_hex(self.to_binary())

    def compute_root(self, txid: Optional[str] = None) -> str:
        """
        Computes the Merkle root from the provided transaction ID.

        Args:
            txid (Optional[str]): The transaction ID to compute the Merkle root for. If not provided,
                the root will be computed from an unspecified branch, and not all branches will be validated!

        Returns:
            str: The computed Merkle root as a hexadecimal string.

        Raises:
            ValueError: If the transaction ID is not part of the Merkle Path.
        """
        if not isinstance(txid, str):
            txid = next(leaf['hash_str'] for leaf in self.path[0] if leaf and 'hash_str' in leaf)

        # Find the index of the txid at the lowest level of the Merkle tree
        try:
            index = next(leaf['offset'] for leaf in self.path[0] if leaf.get('hash_str') == txid)
        except StopIteration:
            raise ValueError(f"This proof does not contain the txid: {txid}")

        # Calculate the root using the index as a way to determine which direction to concatenate.
        def hash_fn(m: str) -> str:
            return to_hex(hash256(to_bytes(m, "hex")[::-1])[::-1])

        working_hash = txid
        for height in range(len(self.path)):
            offset = (index >> height) ^ 1
            leaf = self.find_or_compute_leaf(height, offset)
            if not isinstance(leaf, dict):
                raise ValueError(f"Missing hash for index {index} at height {height}")

            if 'duplicate' in leaf and leaf['duplicate']:
                working_hash = hash_fn(working_hash + working_hash)
            elif offset % 2 != 0:
                working_hash = hash_fn(leaf['hash_str'] + working_hash)
            else:
                working_hash = hash_fn(working_hash + leaf['hash_str'])

        return working_hash

    def find_or_compute_leaf(self, height: int, offset: int) -> Optional[MerkleLeaf]:
        def hash_fn(m: str) -> str:
            return to_hex(hash256(to_bytes(m, "hex")[::-1])[::-1])

        leaf = next((e for e in self.path[height] if e["offset"] == offset), None)
        if leaf:
            return leaf

        if height == 0:
            return None

        h = height - 1
        e = offset << 1

        leaf0 = self.find_or_compute_leaf(h, e)
        if not leaf0 or not leaf0.get("hash_str"):
            return None

        leaf1 = self.find_or_compute_leaf(h, e + 1)
        if not leaf1:
            return None

        if leaf1.get("duplicate"):
            working_hash = hash_fn(leaf0["hash_str"] + leaf0["hash_str"])
        else:
            working_hash = hash_fn(leaf1["hash_str"] + leaf0["hash_str"])

        return {"offset": offset, "hash_str": working_hash}

    async def verify(self, txid: str, chaintracker: ChainTracker) -> bool:
        """
        Verifies if the given transaction ID is part of the Merkle tree at the specified block height.

        Args:
            txid (str): The transaction ID to verify.
            chaintracker (ChainTracker): The ChainTracker instance used to verify the Merkle root.

        Returns:
            bool: True if the transaction ID is valid within the Merkle Path at the specified block height.
        """
        root = self.compute_root(txid)
        res = await chaintracker.is_valid_root_for_height(root, self.block_height)
        return res

    def combine(self, other: "MerklePath") -> None:
        """
        Combines this MerklePath with another to create a compound proof.

        Args:
            other (MerklePath): Another MerklePath to combine with this path.

        Raises:
            ValueError: If the paths have different block heights or roots.
        """
        if self.block_height != other.block_height:
            raise ValueError("You cannot combine paths which do not have the same block height.")

        root1 = self.compute_root()
        root2 = other.compute_root()
        if root1 != root2:
            raise ValueError("You cannot combine paths which do not have the same root.")

        combined_path = []
        for h in range(len(self.path)):
            combined_level = self.path[h] + [
                leaf for leaf in other.path[h] if leaf["offset"] not in {e["offset"] for e in self.path[h]}
            ]
            for leaf in other.path[h]:
                if "txid" in leaf:
                    for e in combined_level:
                        if e["offset"] == leaf["offset"]:
                            e["txid"] = True
            combined_path.append(combined_level)
        self.path = combined_path
        self.trim()

    def trim(self) -> None:
        """
        Remove all internal nodes that are not required by level zero txid nodes.
        Assumes that at least all required nodes are present.
        Leaves all levels sorted by increasing offset.
        """

        def push_if_new(v: int, a: List[int]) -> None:
            if not a or a[-1] != v:
                a.append(v)

        def drop_offsets_from_level(drop_offsets: List[int], level: int) -> None:
            for i in reversed(drop_offsets):
                idx = next((j for j, n in enumerate(self.path[level]) if n["offset"] == i), None)
                if idx is not None:
                    self.path[level].pop(idx)

        def next_computed_offsets(cos: List[int]) -> List[int]:
            ncos = []
            for o in cos:
                push_if_new(o >> 1, ncos)
            return ncos

        computed_offsets = []
        drop_offsets = []
        for h in range(len(self.path)):
            self.path[h].sort(key=lambda x: x["offset"])

        for e in self.path[0]:
            if e.get("txid"):
                push_if_new(e["offset"] >> 1, computed_offsets)
            else:
                is_odd = e["offset"] % 2 == 1
                peer = next((n for n in self.path[0] if n["offset"] == e["offset"] + (1 if is_odd else -1)), None)
                if peer and not peer.get("txid"):
                    push_if_new(peer["offset"], drop_offsets)

        drop_offsets_from_level(drop_offsets, 0)
        print('testing', self.path)
        for h in range(1, len(self.path)):
            drop_offsets = computed_offsets
            computed_offsets = next_computed_offsets(computed_offsets)
            drop_offsets_from_level(drop_offsets, h)



from abc import ABC, abstractmethod
from typing import Union, Dict, Any, TYPE_CHECKING


if TYPE_CHECKING:
    from .transaction import Transaction

class BroadcastResponse:
    def __init__(self, status: str, txid: str, message: str):
        self.status = status
        self.txid = txid
        self.message = message


class BroadcastFailure:
    def __init__(
            self,
            status: str,
            code: str,
            description: str,
            txid: str = None,
            more: Dict[str, Any] = None,
    ):
        self.status = status
        self.code = code
        self.txid = txid
        self.description = description
        self.more = more


class Broadcaster(ABC):
    @abstractmethod
    async def broadcast(
            self, transaction: 'Transaction'
    ) -> Union[BroadcastResponse, BroadcastFailure]:
        pass


def is_broadcast_response(r: Union[BroadcastResponse, BroadcastFailure]) -> bool:
    return r.status == "success"


def is_broadcast_failure(r: Union[BroadcastResponse, BroadcastFailure]) -> bool:
    return r.status == "error"



import hmac
from hashlib import sha512
from typing import Union, List

from .bip39 import seed_from_mnemonic
from ..base58 import base58check_decode, base58check_encode
from ..constants import BIP32_SEED_BYTE_LENGTH
from ..constants import NETWORK_XPUB_PREFIX_DICT, NETWORK_XPRV_PREFIX_DICT
from ..constants import Network, XKEY_BYTE_LENGTH, XKEY_PREFIX_LIST, PUBLIC_KEY_COMPRESSED_PREFIX_LIST
from ..constants import XPUB_PREFIX_NETWORK_DICT, XPRV_PREFIX_NETWORK_DICT, BIP32_DERIVATION_PATH
from ..curve import curve, curve_add, curve_multiply
from ..keys import PublicKey, PrivateKey


class Xkey:
    """
    [  : 4] prefix
    [ 4: 5] depth
    [ 5: 9] parent public key fingerprint
    [ 9:13] child index
    [13:45] chain code
    [45:78] key (private/public)
    """

    def __init__(self, xkey: Union[str, bytes]):
        if isinstance(xkey, str):
            self.payload: bytes = base58check_decode(xkey)
        elif isinstance(xkey, bytes):
            self.payload: bytes = xkey
        else:
            raise TypeError('unsupported extended key type')

        assert len(self.payload) == XKEY_BYTE_LENGTH, 'invalid extended key length'
        self.prefix: bytes = self.payload[:4]
        self.depth: int = self.payload[4]
        self.fingerprint: bytes = self.payload[5:9]
        self.index: int = int.from_bytes(self.payload[9:13], 'big')
        self.chain_code: bytes = self.payload[13:45]
        self.key_bytes: bytes = self.payload[45:]
        assert self.prefix in XKEY_PREFIX_LIST, 'invalid extended key prefix'

    def __eq__(self, o: object) -> bool:
        if isinstance(o, Xkey):
            return self.payload == o.payload
        return super().__eq__(o)  # pragma: no cover

    def __str__(self) -> str:
        return base58check_encode(self.payload)

    def __repr__(self) -> str:  # pragma: no cover
        return self.__str__()


class Xpub(Xkey):

    def __init__(self, xpub: Union[str, bytes]):
        super().__init__(xpub)
        self.network: Network = XPUB_PREFIX_NETWORK_DICT.get(self.prefix)
        assert self.network, 'unknown xpub prefix'
        assert self.payload[45:46] in PUBLIC_KEY_COMPRESSED_PREFIX_LIST, 'invalid public key in xpub'
        self.key: PublicKey = PublicKey(self.key_bytes)

    def ckd(self, index: Union[int, str, bytes]) -> 'Xpub':
        if isinstance(index, int):
            index = index.to_bytes(4, 'big')
        elif isinstance(index, str):
            index = bytes.fromhex(index)
        assert len(index) == 4, 'index should be a 4 bytes integer'
        assert index[0] < 0x80, ("can't make hardened derivation from xpub. "
                                 "If you use hardened key, please set xpub with path from xpriv first. Example:\n"
                                 "  master_xprv = master_xprv_from_seed(seed)\n"
                                 "  account_xprv = ckd(master_xprv, \"m/44'/0'/0'\")\n"
                                 "  account_xpub = account_xprv.xpub()")

        payload: bytes = self.prefix
        payload += (self.depth + 1).to_bytes(1, 'big')
        payload += self.key.hash160()[:4]
        payload += index

        h: bytes = hmac.new(self.chain_code, self.key.serialize() + index, sha512).digest()
        offset: int = int.from_bytes(h[:32], 'big')
        child: PublicKey = PublicKey(curve_add(self.key.point(), curve_multiply(offset, curve.g)))

        payload += h[32:]
        payload += child.serialize()

        return Xpub(payload)

    def public_key(self) -> PublicKey:
        return self.key

    def address(self) -> str:
        return self.key.address(network=self.network)

    @classmethod
    def from_xprv(cls, xprv: Union[str, bytes, 'Xprv']) -> 'Xpub':
        if not isinstance(xprv, Xprv):
            xprv = Xprv(xprv)
        payload: bytes = NETWORK_XPUB_PREFIX_DICT.get(xprv.network)
        payload += xprv.depth.to_bytes(1, 'big')
        payload += xprv.fingerprint
        payload += xprv.index.to_bytes(4, 'big')
        payload += xprv.chain_code
        payload += xprv.key.public_key().serialize()
        return Xpub(payload)


class Xprv(Xkey):

    def __init__(self, xprv: Union[str, bytes]):
        super().__init__(xprv)
        self.network: Network = XPRV_PREFIX_NETWORK_DICT.get(self.prefix)
        assert self.network, 'unknown xprv prefix'
        assert self.payload[45] == 0, 'invalid private key in xprv'
        self.key: PrivateKey = PrivateKey(self.key_bytes[1:], network=self.network)

    def ckd(self, index: Union[int, str, bytes]) -> 'Xprv':
        if isinstance(index, int):
            index = index.to_bytes(4, 'big')
        elif isinstance(index, str):
            index = bytes.fromhex(index)
        assert len(index) == 4, 'index should be a 4 bytes integer'

        payload: bytes = self.prefix
        payload += (self.depth + 1).to_bytes(1, 'big')
        payload += self.key.public_key().hash160()[:4]
        payload += index

        message: bytes = (self.key.public_key().serialize() if index[0] < 0x80 else self.key_bytes) + index
        h: bytes = hmac.new(self.chain_code, message, sha512).digest()
        offset: int = int.from_bytes(h[:32], 'big')
        child: PrivateKey = PrivateKey((self.key.int() + offset) % curve.n)

        payload += h[32:]
        payload += b'\x00' + child.serialize()

        return Xprv(payload)

    def xpub(self) -> Xpub:
        return Xpub.from_xprv(self)

    def private_key(self) -> PrivateKey:
        return self.key

    def public_key(self) -> PublicKey:
        return self.key.public_key()

    def address(self) -> str:
        return self.key.address()

    @classmethod
    def from_seed(cls, seed: Union[str, bytes], network: Network = Network.MAINNET):
        """
        derive master extended private key from seed
        """
        if isinstance(seed, str):
            seed = bytes.fromhex(seed)
        assert len(seed) == BIP32_SEED_BYTE_LENGTH, 'invalid seed byte length'

        payload: bytes = NETWORK_XPRV_PREFIX_DICT.get(network)
        payload += b'\x00'
        payload += b'\x00\x00\x00\x00'
        payload += b'\x00\x00\x00\x00'

        h: bytes = hmac.new(b'Bitcoin seed', seed, sha512).digest()
        payload += h[32:]
        payload += b'\x00' + h[:32]

        return Xprv(payload)


def step_to_index(step: Union[str, int]) -> int:
    """
    convert step (sub path) normal derivation or hardened derivation into child index
    """
    assert type(step).__name__ in ['str', 'int'], 'unsupported step type'
    if isinstance(step, str):
        assert len(step), 'invalid step'
        hardened: bool = step[-1] == "'"
        index: int = (0x80000000 if hardened else 0) + int(step[:-1] if hardened else step)
    else:
        index: int = step
    assert 0 <= index < 0xffffffff, 'step out of range'
    return index


def ckd(xkey: Union[Xprv, Xpub], path: str) -> Union[Xprv, Xpub]:
    """
    ckd = "Child Key Derivation"
    derive an extended key according to path like "m/44'/0'/1'/0/10" (absolute) or "./0/10" (relative)
    """
    steps = path.strip(' ').strip('/').split('/')
    assert steps and steps[0] in ['m', '.']

    if steps[0] == 'm':
        # should be master key
        assert (
                xkey.depth == 0 and xkey.fingerprint == b'\x00\x00\x00\x00' and xkey.index == 0
        ), 'absolute path for non-master key'

    child = xkey
    for step in steps[1:]:
        child = child.ckd(step_to_index(step))
    return child


def master_xprv_from_seed(seed: Union[str, bytes], network: Network = Network.MAINNET) -> Xprv:
    return Xprv.from_seed(seed, network)


def _derive_xkeys_from_xkey(xkey: Union[Xprv, Xpub],
                            index_start: Union[str, int],
                            index_end: Union[str, int],
                            change: Union[str, int] = 0) -> List[Union[Xprv, Xpub]]:
    """
    this function is internal use only within bip32 module
    Use bip32_derive_xkeys_from_xkey instead.
    """
    change_xkey = xkey.ckd(step_to_index(change))
    return [change_xkey.ckd(i) for i in range(step_to_index(index_start), step_to_index(index_end))]


def bip32_derive_xprv_from_mnemonic(mnemonic: str,
                                    lang: str = 'en',
                                    passphrase: str = '',
                                    prefix: str = 'mnemonic',
                                    path: str = BIP32_DERIVATION_PATH,
                                    network: Network = Network.MAINNET) -> Xprv:
    """
    Derive the subtree root extended private key from mnemonic and path.
    """
    seed = seed_from_mnemonic(mnemonic, lang, passphrase, prefix)
    master_xprv = Xprv.from_seed(seed, network)
    return ckd(master_xprv, path)


def bip32_derive_xprvs_from_mnemonic(mnemonic: str,
                                     index_start: Union[str, int],
                                     index_end: Union[str, int],
                                     lang: str = 'en',
                                     passphrase: str = '',
                                     prefix: str = 'mnemonic',
                                     path: str = BIP32_DERIVATION_PATH,
                                     change: Union[str, int] = 0,
                                     network: Network = Network.MAINNET) -> List[Xprv]:
    """
    Derive a range of extended keys from a nmemonic using BIP32 format
    """
    xprv = bip32_derive_xprv_from_mnemonic(mnemonic, lang, passphrase, prefix, path, network)
    return _derive_xkeys_from_xkey(xprv, index_start, index_end, change)


def bip32_derive_xkeys_from_xkey(xkey: Union[Xprv, Xpub],
                                 index_start: Union[str, int],
                                 index_end: Union[str, int],
                                 path: str = BIP32_DERIVATION_PATH,
                                 change: Union[str, int] = 0) -> List[Union[Xprv, Xpub]]:
    """
    Derive a range of extended keys from Xprv and Xpub keys using BIP32 path structure.

    Args:
        xkey: Parent extended key (Xprv or Xpub)
        index_start: Starting index for derivation
        index_end: Ending index for derivation (exclusive)
        path: Base derivation path (default: BIP32_DERIVATION_PATH)
        change: Change level (0 for receiving addresses, 1 for change addresses)

    Returns:
        List[Union[Xprv, Xpub]]: List of derived extended keys
    """
    # Convert index arguments to integers
    start_idx = step_to_index(index_start) if isinstance(index_start, str) else index_start
    end_idx = step_to_index(index_end) if isinstance(index_end, str) else index_end

    # Validate indices
    if start_idx < 0 or end_idx < 0 or start_idx >= end_idx:
        raise ValueError("Invalid index range: start must be non-negative and less than end")

    # Parse the base path and reconstruct with change value
    base_path = path.rstrip('/')  # Remove trailing slashes if any
    if base_path.startswith('m/'):
        # For absolute paths
        derived_path = f"{base_path}/{change}"
    else:
        # For relative paths
        derived_path = f"./{change}"

    # First derive to the change level
    change_level = ckd(xkey, derived_path)

    # Then derive the range of addresses
    derived_keys = []
    for i in range(start_idx, end_idx):
        child_key = change_level.ckd(i)
        derived_keys.append(child_key)

    return derived_keys



import unittest


from bsv.keys import PrivateKey
from bsv.polynomial import KeyShares, PointInFiniteField


class TestPrivateKeySharing(unittest.TestCase):
    # 既知のバックアップシェアデータ
    sample_backup = [
        '45s4vLL2hFvqmxrarvbRT2vZoQYGZGocsmaEksZ64o5M.A7nZrGux15nEsQGNZ1mbfnMKugNnS6SYYEQwfhfbDZG8.3.2f804d43',
        '7aPzkiGZgvU4Jira5PN9Qf9o7FEg6uwy1zcxd17NBhh3.CCt7NH1sPFgceb6phTRkfviim2WvmUycJCQd2BxauxP9.3.2f804d43',
        '9GaS2Tw5sXqqbuigdjwGPwPsQuEFqzqUXo5MAQhdK3es.8MLh2wyE3huyq6hiBXjSkJRucgyKh4jVY6ESq5jNtXRE.3.2f804d43',
        'GBmoNRbsMVsLmEK5A6G28fktUNonZkn9mDrJJ58FXgsf.HDBRkzVUCtZ38ApEu36fvZtDoDSQTv3TWmbnxwwR7kto.3.2f804d43',
        '2gHebXBgPd7daZbsj6w9TPDta3vQzqvbkLtJG596rdN1.E7ZaHyyHNDCwR6qxZvKkPPWWXzFCiKQFentJtvSSH5Bi.3.2f804d43'
    ]

    def test_split_private_key_into_shares_correctly(self):
        """Test that a private key can be split into shares correctly."""
        private_key = PrivateKey()  # Generate random private key
        threshold = 2
        total_shares = 5

        # Split the private key
        shares = private_key.to_key_shares(threshold, total_shares)
        backup = shares.to_backup_format()

        # Check the number of shares
        self.assertEqual(len(backup), total_shares)

        # Check that each share is a PointInFiniteField
        for share in shares.points:
            self.assertIsInstance(share, PointInFiniteField)

        # Check the threshold
        self.assertEqual(shares.threshold, threshold)

    def test_recombine_shares_into_private_key_correctly(self):
        """Test that shares can be recombined to recover the original key."""
        for _ in range(3):
            key = PrivateKey()
            all_shares = key.to_key_shares(3, 5)
            backup = all_shares.to_backup_format()

            # Use only the first 3 shares (the threshold)
            some_shares = KeyShares.from_backup_format(backup[:3])
            rebuilt_key = PrivateKey.from_key_shares(some_shares)

            # Check if the recovered key matches the original
            self.assertEqual(rebuilt_key.wif(), key.wif())

    def test_invalid_threshold_or_total_shares_type(self):
        """Test that invalid threshold or totalShares types raise errors."""
        k = PrivateKey()

        # Test with invalid threshold type
        with self.assertRaises(ValueError) as cm:
            k.to_key_shares("invalid", 14)  # type: ignore
        self.assertIn("threshold and totalShares must be numbers", str(cm.exception))

        # Test with invalid totalShares type
        with self.assertRaises(ValueError) as cm:
            k.to_key_shares(4, None)  # type: ignore
        self.assertIn("threshold and totalShares must be numbers", str(cm.exception))

    def test_invalid_threshold_value(self):
        """Test that invalid threshold values raise errors."""
        k = PrivateKey()

        # Test with threshold less than 2
        with self.assertRaises(ValueError) as cm:
            k.to_key_shares(1, 2)
        self.assertIn("threshold must be at least 2", str(cm.exception))

    def test_invalid_total_shares_value(self):
        """Test that invalid totalShares values raise errors."""
        k = PrivateKey()

        # Test with negative totalShares
        with self.assertRaises(ValueError) as cm:
            k.to_key_shares(2, -4)
        self.assertIn("totalShares must be at least 2", str(cm.exception))

    def test_threshold_greater_than_total_shares(self):
        """Test that threshold greater than totalShares raises an error."""
        k = PrivateKey()

        # Test with threshold > totalShares
        with self.assertRaises(ValueError) as cm:
            k.to_key_shares(3, 2)
        self.assertIn("threshold should be less than or equal to totalShares", str(cm.exception))

    def test_duplicate_share_in_recovery_with_sample_data(self):
        """Test that using duplicate shares from sample data during recovery raises an error."""
        # 既知のバックアップデータから重複するシェアを含むリストを作成
        duplicate_shares = [
            self.sample_backup[0],
            self.sample_backup[1],
            self.sample_backup[1]  # 重複するシェア
        ]

        # KeySharesオブジェクトを作成
        recovery = KeyShares.from_backup_format(duplicate_shares)

        # 重複するシェアがあるため、キーの復元時にエラーが発生することを確認
        with self.assertRaises(ValueError) as cm:
            PrivateKey.from_key_shares(recovery)
        self.assertIn("Duplicate share detected, each must be unique", str(cm.exception))

    def test_parse_and_verify_sample_shares(self):
        """Test parsing and verification of sample backup shares."""
        # サンプルバックアップデータからKeySharesオブジェクトを作成
        shares = KeyShares.from_backup_format(self.sample_backup[:3])

        # 基本的な検証
        self.assertEqual(shares.threshold, 3)
        self.assertEqual(shares.integrity, "2f804d43")
        self.assertEqual(len(shares.points), 3)

        # 各ポイントがPointInFiniteFieldインスタンスであることを確認
        for point in shares.points:
            self.assertIsInstance(point, PointInFiniteField)

        # バックアップ形式に戻せることを確認
        backup_format = shares.to_backup_format()
        self.assertEqual(len(backup_format), 3)

        # 元のバックアップと同じフォーマットであることを確認
        for i in range(3):
            parts_original = self.sample_backup[i].split('.')
            parts_new = backup_format[i].split('.')

            # 最後の2つの部分（しきい値と整合性ハッシュ）が同じか確認
            self.assertEqual(parts_original[-2:], parts_new[-2:])

    def test_recombination_with_sample_shares(self):
        """Test recombination of private key using different combinations of sample shares."""
        # サンプルシェアの様々な組み合わせでキーを復元
        combinations = [
            [0, 1, 2],  # 最初の3つのシェア
            [0, 2, 4],  # 異なる3つのシェア
            [1, 3, 4]  # 別の組み合わせ
        ]

        # 各組み合わせでキーを復元
        for combo in combinations:
            selected_shares = [self.sample_backup[i] for i in combo]
            key_shares = KeyShares.from_backup_format(selected_shares)

            # キーを復元（例外が投げられなければテストは成功）
            recovered_key = PrivateKey.from_key_shares(key_shares)

            # 復元されたキーがPrivateKeyインスタンスであることを確認
            self.assertIsInstance(recovered_key, PrivateKey)

            # WIFを生成できることを確認
            wif = recovered_key.wif()
            self.assertIsInstance(wif, str)
            self.assertTrue(len(wif) > 0)

    def test_create_backup_and_recover(self):
        """Test creating backup shares and recovering the key from them."""
        key = PrivateKey()
        backup = key.to_backup_shares(3, 5)

        # Recover using only the first 3 shares
        recovered_key = PrivateKey.from_backup_shares(backup[:3])

        # Verify the recovered key matches the original
        self.assertEqual(recovered_key.wif(), key.wif())

    def test_insufficient_shares_for_recovery(self):
        """Test that attempting to recover with insufficient shares raises an error."""
        key = PrivateKey()
        all_shares = key.to_key_shares(3, 5)
        backup = all_shares.to_backup_format()

        # しきい値未満のシェアでKeySharesオブジェクトを作成
        insufficient_shares = KeyShares.from_backup_format(backup[:2])

        # シェアが不足しているため、キーの復元時にエラーが発生することを確認
        with self.assertRaises(ValueError) as cm:
            PrivateKey.from_key_shares(insufficient_shares)
        self.assertIn("At least 3 shares are required", str(cm.exception))

    def test_share_format_validation(self):
        """Test validation of share format."""
        # 不正なフォーマットのシェア
        invalid_shares = [
            '45s4vLL2hFvqmxrarvbRT2vZoQYGZGocsmaEksZ64o5M.A7nZrGux15nEsQGNZ1mbfnMKugNnS6SYYEQwfhfbDZG8.3',  # 完全ではない
            'invalid-format',  # 完全に無効
            '45s4vLL2hFvqmxrarvbRT2vZoQYGZGocsmaEksZ64o5M'  # ドットがない
        ]

        # 各無効なシェアに対して、エラーが発生することを確認
        for invalid_share in invalid_shares:
            with self.assertRaises(ValueError):
                KeyShares.from_backup_format([invalid_share])


if __name__ == '__main__':
    unittest.main()


from .hash import hash256

BASE58_ALPHABET = '123456789ABCDEFGHJKLMNPQRSTUVWXYZabcdefghijkmnopqrstuvwxyz'


def _checksum(payload: bytes) -> bytes:
    return hash256(payload)[:4]


def b58_encode(payload: bytes) -> str:
    pad = 0
    for byte in payload:
        if byte == 0:
            pad += 1
        else:
            break
    prefix = '1' * pad
    num = int.from_bytes(payload, 'big')
    result = ''
    while num > 0:
        num, remaining = divmod(num, 58)
        result = BASE58_ALPHABET[remaining] + result
    return prefix + result


def base58check_encode(payload: bytes) -> str:
    return b58_encode(payload + _checksum(payload))


def to_base58check(payload: bytes, prefix: bytes) -> str:
    """
    Converts a binary array into a base58check string with a checksum
    :param payload: The binary array to convert to base58check
    :param prefix: The prefix to add to the binary
    :return: The base58check string representation
    """
    return base58check_encode(prefix + payload)


def from_base58check(encoded: str, prefix_len: int = 1) -> (bytes, bytes):
    """
    Converts a base58check string into payload and prefix
    :param encoded: The base58check string to convert
    :param prefix_len: The byte length of the prefix
    :return: A tuple containing the prefix and the payload
    """
    payload = base58check_decode(encoded)
    return payload[:prefix_len], payload[prefix_len:]


def b58_decode(encoded: str) -> bytes:
    pad = 0
    for char in encoded:
        if char == '1':
            pad += 1
        else:
            break
    prefix = b'\x00' * pad
    num = 0
    try:
        for char in encoded:
            num *= 58
            num += BASE58_ALPHABET.index(char)
    except Exception:
        raise ValueError(f'invalid base58 encoded {encoded}')
    # if num is 0 then (0).to_bytes will return b''
    return prefix + num.to_bytes((num.bit_length() + 7) // 8, 'big')


def base58check_decode(encoded: str) -> bytes:
    decoded = b58_decode(encoded)
    payload = decoded[:-4]
    decoded_checksum = decoded[-4:]
    hash_checksum = _checksum(payload)
    if decoded_checksum != hash_checksum:
        _msg = f'unmatched base58 checksum, expect {decoded_checksum.hex()} but actually {hash_checksum.hex()}'
        raise ValueError(_msg)
    return payload



import hashlib

import ecdsa
import pytest

from bsv.constants import Network
from bsv.curve import Point
from bsv.hash import sha256
from bsv.keys import PrivateKey, PublicKey, verify_signed_text
from bsv.utils import text_digest, unstringify_ecdsa_recoverable
from .test_transaction import digest1, digest2, digest3

private_key_hex = 'f97c89aaacf0cd2e47ddbacc97dae1f88bec49106ac37716c451dcdd008a4b62'
private_key_bytes = bytes.fromhex(private_key_hex)
private_key_int = int(private_key_hex, 16)
private_key = PrivateKey(private_key_int)

x = 'e46dcd7991e5a4bd642739249b0158312e1aee56a60fd1bf622172ffe65bd789'
y = '97693d32c540ac253de7a3dc73f7e4ba7b38d2dc1ecc8e07920b496fb107d6b2'
point = Point(int(x, 16), int(y, 16))
public_key = PublicKey(point)

address_compressed_main = '1AfxgwYJrBgriZDLryfyKuSdBsi59jeBX9'
address_uncompressed_main = '1BVHzn1J8VZWRuVWbPrj2Szx1j7hHdt5zP'
address_compressed_test = 'mqBuyzdHfD87VfgxaYeM9pex3sJn4ihYHY'
address_uncompressed_test = 'mr1FHq6GwWzmD1y8Jxq6rNDGsiiQ9caF7r'


def test_public_key():
    public_key_compressed = f'02{x}'
    public_key_uncompressed = f'04{x}{y}'

    assert public_key.point() == point
    assert public_key.hex() == public_key_compressed
    assert public_key.hex(compressed=True) == public_key_compressed
    assert public_key.hex(compressed=False) == public_key_uncompressed

    assert public_key.address() == address_compressed_main
    assert public_key.address(compressed=True, network=Network.MAINNET) == address_compressed_main
    assert public_key.address(compressed=False, network=Network.MAINNET) == address_uncompressed_main
    assert public_key.address(compressed=True, network=Network.TESTNET) == address_compressed_test
    assert public_key.address(compressed=False, network=Network.TESTNET) == address_uncompressed_test

    assert PublicKey(public_key_compressed) == public_key
    assert PublicKey(public_key_compressed).address() == address_compressed_main

    assert PublicKey(public_key_uncompressed) == public_key
    assert PublicKey(public_key_uncompressed).address() == address_uncompressed_main

    assert PublicKey(bytes.fromhex(public_key_compressed)) == public_key

    with pytest.raises(TypeError, match=r'unsupported public key type'):
        # noinspection PyTypeChecker
        PublicKey(1.23)


def test_private_key():
    assert private_key == PrivateKey.from_hex(private_key_hex)
    assert private_key.public_key() == public_key
    assert private_key.hex() == private_key_hex
    assert private_key.serialize() == private_key_bytes
    assert private_key.int() == private_key_int

    priv_key_wif_compressed_main = 'L5agPjZKceSTkhqZF2dmFptT5LFrbr6ZGPvP7u4A6dvhTrr71WZ9'
    priv_key_wif_uncompressed_main = '5KiANv9EHEU4o9oLzZ6A7z4xJJ3uvfK2RLEubBtTz1fSwAbpJ2U'
    priv_key_wif_compressed_test = 'cVwfreZB3i8iv9JpdSStd9PWhZZGGJCFLS4rEKWfbkahibwhticA'
    priv_key_wif_uncompressed_test = '93UnxexmsTYCmDJdctz4zacuwxQd5prDmH6rfpEyKkQViAVA3me'

    assert private_key.wif() == priv_key_wif_compressed_main
    assert private_key.wif(compressed=True, network=Network.MAINNET) == priv_key_wif_compressed_main
    assert private_key.wif(compressed=False, network=Network.MAINNET) == priv_key_wif_uncompressed_main
    assert private_key.wif(compressed=True, network=Network.TESTNET) == priv_key_wif_compressed_test
    assert private_key.wif(compressed=False, network=Network.TESTNET) == priv_key_wif_uncompressed_test

    assert PrivateKey(private_key_bytes) == private_key
    assert PrivateKey(priv_key_wif_compressed_main) == private_key
    assert PrivateKey(priv_key_wif_uncompressed_main) == private_key
    assert PrivateKey(priv_key_wif_compressed_test) == private_key
    assert PrivateKey(priv_key_wif_uncompressed_test) == private_key

    assert PrivateKey(private_key_bytes).wif() == priv_key_wif_compressed_main
    assert PrivateKey(private_key_bytes).address() == address_compressed_main

    assert PrivateKey(priv_key_wif_compressed_main).wif() == priv_key_wif_compressed_main
    assert PrivateKey(priv_key_wif_compressed_main).address() == address_compressed_main

    assert PrivateKey(priv_key_wif_uncompressed_main).wif() == priv_key_wif_uncompressed_main
    assert PrivateKey(priv_key_wif_uncompressed_main).address() == address_uncompressed_main

    assert PrivateKey(priv_key_wif_compressed_test).wif() == priv_key_wif_compressed_test
    assert PrivateKey(priv_key_wif_compressed_test).address() == address_compressed_test

    assert PrivateKey(priv_key_wif_uncompressed_test).wif() == priv_key_wif_uncompressed_test
    assert PrivateKey(priv_key_wif_uncompressed_test).address() == address_uncompressed_test

    with pytest.raises(TypeError, match=r'unsupported private key type'):
        # noinspection PyTypeChecker
        PrivateKey(1.23)


def test_verify():
    # https://whatsonchain.com/tx/4674da699de44c9c5d182870207ba89e5ccf395e5101dab6b0900bbf2f3b16cb
    der: bytes = bytes.fromhex('304402207e2c6eb8c4b20e251a71c580373a2836e209c50726e5f8b0f4f59f8af00eee1a'
                               '022019ae1690e2eb4455add6ca5b86695d65d3261d914bc1d7abb40b188c7f46c9a5')
    assert private_key.verify(der, digest1)

    # https://whatsonchain.com/tx/c04bbd007ad3987f9b2ea8534175b5e436e43d64471bf32139b5851adf9f477e
    der: bytes = bytes.fromhex('3043022053b1f5a28a011c60614401eeef88e49c676a098ce36d95ded1b42667f40efa37'
                               '021f4de6703f8c74b0ce5dad617c00d1fb99580beb7972bf681e7215911c3648de')
    assert private_key.verify(der, digest2)
    der: bytes = bytes.fromhex('3045022100b9f293781ae1e269591df779dbadb41b9971d325d7b8f83d883fb55f2cb3ff76'
                               '02202fe1e822628d85b0f52966602d0e153be411980d54884fa48a41d6fc32b4e9f5')
    assert private_key.verify(der, digest3)


def test_sign():
    # ecdsa
    message: bytes = b'hello world'
    der: bytes = private_key.sign(message)
    vk = ecdsa.VerifyingKey.from_string(public_key.serialize(), curve=ecdsa.SECP256k1)
    assert vk.verify(signature=der, data=sha256(message), hashfunc=hashlib.sha256, sigdecode=ecdsa.util.sigdecode_der)

    # recoverable ecdsa
    text = 'hello world'
    address, signature = private_key.sign_text(text)
    assert verify_signed_text(text, address, signature)

    message: bytes = text_digest(text)
    serialized_recoverable, _ = unstringify_ecdsa_recoverable(signature)
    assert private_key.verify_recoverable(serialized_recoverable, message)

    address, signature = PrivateKey('5KiANv9EHEU4o9oLzZ6A7z4xJJ3uvfK2RLEubBtTz1fSwAbpJ2U').sign_text(text)
    assert verify_signed_text(text, address, signature)


def test_ecdh():
    alice, bob = PrivateKey(), PrivateKey()
    assert alice.derive_shared_secret(bob.public_key()) == bob.derive_shared_secret(alice.public_key())
    ephemeral = PrivateKey()
    assert alice.public_key().derive_shared_secret(ephemeral) == alice.derive_shared_secret(ephemeral.public_key())


def test_encryption():
    plain = 'hello world'
    encrypted = ('QklFMQPkjNG3xxnfRv7oUDjUYPH2VN3VFrcglCcwmeYpJpsjRKnfl/XsS+dOg'
                 'ocRV6JKVHkfUZAKIHDo7vwxjv/BPkV5EA2Dl4RJ6d/jpWwgGdFBYA==')
    assert private_key.decrypt_text(encrypted) == plain
    assert private_key.decrypt_text(public_key.encrypt_text(plain)) == plain


def test_brc42():
    # https://github.com/bitcoin-sv/BRCs/blob/master/key-derivation/0042.md#test-vectors
    private_key_derivation_cases = [{
        'senderPublicKey': '033f9160df035156f1c48e75eae99914fa1a1546bec19781e8eddb900200bff9d1',
        'recipientPrivateKey': '6a1751169c111b4667a6539ee1be6b7cd9f6e9c8fe011a5f2fe31e03a15e0ede',
        'invoiceNumber': 'f3WCaUmnN9U=',
        'privateKey': '761656715bbfa172f8f9f58f5af95d9d0dfd69014cfdcacc9a245a10ff8893ef'
    }, {
        'senderPublicKey': '027775fa43959548497eb510541ac34b01d5ee9ea768de74244a4a25f7b60fae8d',
        'recipientPrivateKey': 'cab2500e206f31bc18a8af9d6f44f0b9a208c32d5cca2b22acfe9d1a213b2f36',
        'invoiceNumber': '2Ska++APzEc=',
        'privateKey': '09f2b48bd75f4da6429ac70b5dce863d5ed2b350b6f2119af5626914bdb7c276'
    }, {
        'senderPublicKey': '0338d2e0d12ba645578b0955026ee7554889ae4c530bd7a3b6f688233d763e169f',
        'recipientPrivateKey': '7a66d0896f2c4c2c9ac55670c71a9bc1bdbdfb4e8786ee5137cea1d0a05b6f20',
        'invoiceNumber': 'cN/yQ7+k7pg=',
        'privateKey': '7114cd9afd1eade02f76703cc976c241246a2f26f5c4b7a3a0150ecc745da9f0'
    }, {
        'senderPublicKey': '02830212a32a47e68b98d477000bde08cb916f4d44ef49d47ccd4918d9aaabe9c8',
        'recipientPrivateKey': '6e8c3da5f2fb0306a88d6bcd427cbfba0b9c7f4c930c43122a973d620ffa3036',
        'invoiceNumber': 'm2/QAsmwaA4=',
        'privateKey': 'f1d6fb05da1225feeddd1cf4100128afe09c3c1aadbffbd5c8bd10d329ef8f40'
    }, {
        'senderPublicKey': '03f20a7e71c4b276753969e8b7e8b67e2dbafc3958d66ecba98dedc60a6615336d',
        'recipientPrivateKey': 'e9d174eff5708a0a41b32624f9b9cc97ef08f8931ed188ee58d5390cad2bf68e',
        'invoiceNumber': 'jgpUIjWFlVQ=',
        'privateKey': 'c5677c533f17c30f79a40744b18085632b262c0c13d87f3848c385f1389f79a6'
    }]
    for case in private_key_derivation_cases:
        sender_public_key = PublicKey(case['senderPublicKey'])
        recipient_private_key = PrivateKey.from_hex(case['recipientPrivateKey'])
        invoice_number = case['invoiceNumber']
        correct_private_key = case['privateKey']
        assert recipient_private_key.derive_child(sender_public_key, invoice_number).hex() == correct_private_key

    public_key_derivation_cases = [{
        'senderPrivateKey': '583755110a8c059de5cd81b8a04e1be884c46083ade3f779c1e022f6f89da94c',
        'recipientPublicKey': '02c0c1e1a1f7d247827d1bcf399f0ef2deef7695c322fd91a01a91378f101b6ffc',
        'invoiceNumber': 'IBioA4D/OaE=',
        'publicKey': '03c1bf5baadee39721ae8c9882b3cf324f0bf3b9eb3fc1b8af8089ca7a7c2e669f'
    }, {
        'senderPrivateKey': '2c378b43d887d72200639890c11d79e8f22728d032a5733ba3d7be623d1bb118',
        'recipientPublicKey': '039a9da906ecb8ced5c87971e9c2e7c921e66ad450fd4fc0a7d569fdb5bede8e0f',
        'invoiceNumber': 'PWYuo9PDKvI=',
        'publicKey': '0398cdf4b56a3b2e106224ff3be5253afd5b72de735d647831be51c713c9077848'
    }, {
        'senderPrivateKey': 'd5a5f70b373ce164998dff7ecd93260d7e80356d3d10abf928fb267f0a6c7be6',
        'recipientPublicKey': '02745623f4e5de046b6ab59ce837efa1a959a8f28286ce9154a4781ec033b85029',
        'invoiceNumber': 'X9pnS+bByrM=',
        'publicKey': '0273eec9380c1a11c5a905e86c2d036e70cbefd8991d9a0cfca671f5e0bbea4a3c'
    }, {
        'senderPrivateKey': '46cd68165fd5d12d2d6519b02feb3f4d9c083109de1bfaa2b5c4836ba717523c',
        'recipientPublicKey': '031e18bb0bbd3162b886007c55214c3c952bb2ae6c33dd06f57d891a60976003b1',
        'invoiceNumber': '+ktmYRHv3uQ=',
        'publicKey': '034c5c6bf2e52e8de8b2eb75883090ed7d1db234270907f1b0d1c2de1ddee5005d'
    }, {
        'senderPrivateKey': '7c98b8abd7967485cfb7437f9c56dd1e48ceb21a4085b8cdeb2a647f62012db4',
        'recipientPublicKey': '03c8885f1e1ab4facd0f3272bb7a48b003d2e608e1619fb38b8be69336ab828f37',
        'invoiceNumber': 'PPfDTTcl1ao=',
        'publicKey': '03304b41cfa726096ffd9d8907fe0835f888869eda9653bca34eb7bcab870d3779'
    }]
    for case in public_key_derivation_cases:
        sender_private_key = PrivateKey.from_hex(case['senderPrivateKey'])
        recipient_public_key = PublicKey(case['recipientPublicKey'])
        invoice_number = case['invoiceNumber']
        correct_public_key = case['publicKey']
        assert recipient_public_key.derive_child(sender_private_key, invoice_number).hex() == correct_public_key



from typing import Union, TYPE_CHECKING

from ..broadcaster import Broadcaster, BroadcastFailure, BroadcastResponse
from ..http_client import HttpClient, default_http_client
from ..constants import Network

if TYPE_CHECKING:
    from ..transaction import Transaction

class WhatsOnChainBroadcaster(Broadcaster):
    def __init__(self, network: Union[Network, str] = Network.MAINNET, http_client: HttpClient = None):
        """
        Initialize WhatsOnChainBroadcaster.
        
        :param network: Network to broadcast to. Can be either Network enum or string ('main'/'test')
        :param http_client: Optional HTTP client to use for requests
        """
        if isinstance(network, str):
            network_str = network.lower()
            if network_str in ['main', 'mainnet']:
                self.network = 'main'
            elif network_str in ['test', 'testnet']:
                self.network = 'test'
            else:
                raise ValueError(f"Invalid network string: {network}. Must be 'main' or 'test'")
        else:
            self.network = 'main' if network == Network.MAINNET else 'test'
            
        self.URL = f"https://api.whatsonchain.com/v1/bsv/{self.network}/tx/raw"
        self.http_client = http_client if http_client else default_http_client()

    async def broadcast(
        self, tx: 'Transaction'
    ) -> Union[BroadcastResponse, BroadcastFailure]:
        request_options = {
            "method": "POST",
            "headers": {"Content-Type": "application/json", "Accept": "text/plain"},
            "data": {"txhex": tx.hex()},
        }

        try:
            response = await self.http_client.fetch(self.URL, request_options)
            if response.ok:
                txid = response.json()["data"]
                return BroadcastResponse(
                    status="success", txid=txid, message="broadcast successful"
                )
            else:
                return BroadcastFailure(
                    status="error",
                    code=str(response.status_code),
                    description=response.json()["data"],
                )
        except Exception as error:
            return BroadcastFailure(
                status="error",
                code="500",
                description=(str(error) if str(error) else "Internal Server Error"),
            )



import pytest
from bsv.broadcasters.whatsonchain import WhatsOnChainBroadcaster
from bsv.constants import Network
from bsv.broadcaster import BroadcastResponse, BroadcastFailure


class TestWhatsOnChainBroadcast:
   def test_network_enum(self):
       # Initialize with Network enum
       broadcaster = WhatsOnChainBroadcaster(Network.MAINNET)
       assert broadcaster.URL == "https://api.whatsonchain.com/v1/bsv/main/tx/raw"

       broadcaster = WhatsOnChainBroadcaster(Network.TESTNET)
       assert broadcaster.URL == "https://api.whatsonchain.com/v1/bsv/test/tx/raw"

   def test_network_string(self):
       # Initialize with string (backward compatibility)
       broadcaster = WhatsOnChainBroadcaster("main")
       assert broadcaster.URL == "https://api.whatsonchain.com/v1/bsv/main/tx/raw"

       broadcaster = WhatsOnChainBroadcaster("test")
       assert broadcaster.URL == "https://api.whatsonchain.com/v1/bsv/test/tx/raw"

       broadcaster = WhatsOnChainBroadcaster("mainnet")
       assert broadcaster.URL == "https://api.whatsonchain.com/v1/bsv/main/tx/raw"

       broadcaster = WhatsOnChainBroadcaster("testnet")
       assert broadcaster.URL == "https://api.whatsonchain.com/v1/bsv/test/tx/raw"

   def test_invalid_network(self):
       # Test invalid network string
       with pytest.raises(ValueError, match="Invalid network string:"):
           WhatsOnChainBroadcaster("invalid_network")


import pytest

from bsv.base58 import base58check_encode, b58_encode
from bsv.constants import Network, OpCode
from bsv.curve import curve
from bsv.utils import bytes_to_bits, bits_to_bytes
from bsv.utils import decode_address, address_to_public_key_hash, decode_wif, validate_address
from bsv.utils import get_pushdata_code, encode_pushdata, encode_int
from bsv.utils import serialize_ecdsa_recoverable, deserialize_ecdsa_recoverable
from bsv.utils import stringify_ecdsa_recoverable, unstringify_ecdsa_recoverable
from bsv.utils import text_digest
from bsv.utils import unsigned_to_varint, unsigned_to_bytes, deserialize_ecdsa_der, serialize_ecdsa_der


def test_unsigned_to_varint():
    assert unsigned_to_varint(0) == bytes.fromhex('00')
    assert unsigned_to_varint(0xfc) == bytes.fromhex('fc')

    assert unsigned_to_varint(0xfd) == bytes.fromhex('fdfd00')
    assert unsigned_to_varint(0xabcd) == bytes.fromhex('fdcdab')

    assert unsigned_to_varint(0x010000) == bytes.fromhex('fe00000100')
    assert unsigned_to_varint(0x12345678) == bytes.fromhex('fe78563412')

    assert unsigned_to_varint(0x0100000000) == bytes.fromhex('ff0000000001000000')
    assert unsigned_to_varint(0x1234567890abcdef) == bytes.fromhex('ffefcdab9078563412')

    with pytest.raises(OverflowError):
        unsigned_to_varint(-1)
    with pytest.raises(OverflowError):
        unsigned_to_varint(0x010000000000000000)


def test_unsigned_to_bytes():
    with pytest.raises(OverflowError):
        unsigned_to_bytes(-1)

    assert unsigned_to_bytes(0) == bytes.fromhex('00')
    assert unsigned_to_bytes(num=255, byteorder='big') == bytes.fromhex('ff')
    assert unsigned_to_bytes(num=256, byteorder='big') == bytes.fromhex('0100')

    assert unsigned_to_bytes(num=256, byteorder='little') == bytes.fromhex('0001')


def test_address():
    a1 = '1A1zP1eP5QGefi2DMPTfTL5SLmv7DivfNa'
    pkh1 = bytes.fromhex('62e907b15cbf27d5425399ebf6f0fb50ebb88f18')
    assert decode_address(a1) == (pkh1, Network.MAINNET)

    a2 = 'moEoqh2ZfYU8jN5EG6ERw6E3DmwnkuTdBC'
    pkh2 = bytes.fromhex('54b34b1ba228ba1d75dca5a40a114dc0f13a2687')
    assert decode_address(a2) == (pkh2, Network.TESTNET)

    a3 = 'n34P4t4K6bJtc6qfGU2pqcRix8mUACdNyJ'
    pkh3 = bytes.fromhex('ec4c3733cff428e9a3c1434274b109fbe2a33b62')
    assert address_to_public_key_hash(a3) == pkh3

    address_invalid_prefix = base58check_encode(b'\xff' + bytes.fromhex('62e907b15cbf27d5425399ebf6f0fb50ebb88f18'))
    with pytest.raises(ValueError, match=r'invalid P2PKH address'):
        decode_address(address_invalid_prefix)

    address_invalid_checksum = b58_encode(b'\x00' + bytes.fromhex('62e907b15cbf27d5425399ebf6f0fb50ebb88f18') + b'\x00')
    with pytest.raises(ValueError, match=r'unmatched base58 checksum'):
        decode_address(address_invalid_checksum)

    assert validate_address('1A1zP1eP5QGefi2DMPTfTL5SLmv7DivfNa')
    assert validate_address('1A1zP1eP5QGefi2DMPTfTL5SLmv7DivfNa', Network.MAINNET)
    assert not validate_address('1A1zP1eP5QGefi2DMPTfTL5SLmv7DivfNa', Network.TESTNET)
    assert validate_address('moEoqh2ZfYU8jN5EG6ERw6E3DmwnkuTdBC', Network.TESTNET)
    assert not validate_address('moEoqh2ZfYU8jN5EG6ERw6E3DmwnkuTdB')
    assert not validate_address('')
    assert not validate_address(address_invalid_prefix)
    assert not validate_address(address_invalid_checksum)


def test_decode_wif():
    private_key_bytes = bytes.fromhex('f97c89aaacf0cd2e47ddbacc97dae1f88bec49106ac37716c451dcdd008a4b62')
    wif_compressed_main = 'L5agPjZKceSTkhqZF2dmFptT5LFrbr6ZGPvP7u4A6dvhTrr71WZ9'
    wif_uncompressed_main = '5KiANv9EHEU4o9oLzZ6A7z4xJJ3uvfK2RLEubBtTz1fSwAbpJ2U'
    wif_compressed_test = 'cVwfreZB3i8iv9JpdSStd9PWhZZGGJCFLS4rEKWfbkahibwhticA'
    wif_uncompressed_test = '93UnxexmsTYCmDJdctz4zacuwxQd5prDmH6rfpEyKkQViAVA3me'

    assert decode_wif(wif_compressed_main) == (private_key_bytes, True, Network.MAINNET)
    assert decode_wif(wif_uncompressed_main) == (private_key_bytes, False, Network.MAINNET)
    assert decode_wif(wif_compressed_test) == (private_key_bytes, True, Network.TESTNET)
    assert decode_wif(wif_uncompressed_test) == (private_key_bytes, False, Network.TESTNET)

    with pytest.raises(ValueError, match=r'unknown WIF prefix'):
        decode_wif(base58check_encode(b'\xff' + private_key_bytes))


def test_der_serialization():
    der1: str = ('3045022100fd5647a062d42cdde975ad4796cefd6b5613e731c08e0fb6907f757a60f44b02'
                 '0220350fee392713423ebfcd8026ea29cc95917d823392f07cd6c80f46712650388e')
    r1 = 114587593887127314608220924841831336233967095853165151956820984900193959037698
    s1 = 24000727837347392504013031837120627225728348681623127776947626422811445180558

    der2: str = ('304402207e2c6eb8c4b20e251a71c580373a2836e209c50726e5f8b0f4f59f8af00eee1a'
                 '022019ae1690e2eb4455add6ca5b86695d65d3261d914bc1d7abb40b188c7f46c9a5')
    r2 = 57069924365784604413146650701306419944030991562754207986153667089859857018394
    s2 = 11615408348402409164215774430388304177694127390766203039231142052414850779557

    der3: str = ('3044022023f093813911a658ac7cbaeb8ba7828b4067ea3582c78f8bd2c38b1f317489ba'
                 '022000e1e43145a89f0d9d8524798b8ae2ca60ebf3947e35106d5e1ddf398985a033')
    r3 = 16256011036517295435281672405882454685603286080662722236323812471789728336314
    s3 = 399115516115506318232804590771004057701078428754012727453057145885291814963

    assert serialize_ecdsa_der((r1, s1)).hex() == der1
    assert serialize_ecdsa_der((r1, curve.n - s1)).hex() == der1
    assert serialize_ecdsa_der((r2, s2)).hex() == der2
    assert serialize_ecdsa_der((r2, curve.n - s2)).hex() == der2
    assert serialize_ecdsa_der((r3, s3)).hex() == der3
    assert serialize_ecdsa_der((r3, curve.n - s3)).hex() == der3

    assert deserialize_ecdsa_der(bytes.fromhex(der1)) == (r1, s1)
    assert deserialize_ecdsa_der(bytes.fromhex(der2)) == (r2, s2)
    with pytest.raises(ValueError, match=r'invalid DER encoded'):
        deserialize_ecdsa_der(b'')


def test_recoverable_serialization():
    sig1 = 'IGdzMq98lowek10e3JFXWj909xp0oLRj71aF7jpWRxaabwH+fBia/K2JpoGQlFFbAl/Q5jo2DYSzQw6pZWhmRtk='
    r1 = 46791760634954614230959036903197650877536710453529507613159894982805988775578
    s1 = 50210249429004071986853078788876176203428035162933045037212292756431067039449
    rec1 = 1
    serialized1, compressed1 = unstringify_ecdsa_recoverable(sig1)
    assert compressed1
    assert serialize_ecdsa_recoverable((r1, s1, rec1)) == serialized1
    assert deserialize_ecdsa_recoverable(serialized1) == (r1, s1, rec1)
    assert stringify_ecdsa_recoverable(serialized1, compressed1) == sig1

    sig2 = 'G1CbjucJgMF/5lyS7LPZrLZPVU60RA6b7fq9b1zULG6uNq4PWQUD8HAvZMgKRPk/vkbDwN0ZsPwoVgKgV5rOSyI='
    r2 = 36459875458431662725541158294877706686723420026424146605771954142876183326382
    s2 = 24732431138926461036459634608851410023678722603615132417233328850542638549794
    rec2 = 0
    serialized2, compressed2 = unstringify_ecdsa_recoverable(sig2)
    assert not compressed2
    assert serialize_ecdsa_recoverable((r2, s2, rec2)) == serialized2
    assert deserialize_ecdsa_recoverable(serialized2) == (r2, s2, rec2)
    assert stringify_ecdsa_recoverable(serialized2, compressed2) == sig2


def test_text_digest():
    message = 'hello world'
    assert text_digest(message).hex() == '18426974636f696e205369676e6564204d6573736167653a0a0b68656c6c6f20776f726c64'


def test_bits():
    assert bytes_to_bits(b'\x00') == '00000000'
    assert bytes_to_bits('12') == '00010010'
    assert bytes_to_bits('f1') == '11110001'
    assert bytes_to_bits('0001') == '0000000000000001'

    assert bits_to_bytes('101') == b'\x05'
    assert bits_to_bytes('100010101010111') == b'\x45\x57'
    assert bits_to_bytes('000000000000001') == b'\x00\x01'
    assert bits_to_bytes('0000000000000001') == b'\x00\x01'


def test_get_pushdata_code():
    assert get_pushdata_code(0x4b) == b'\x4b'
    assert get_pushdata_code(0x4c) == bytes.fromhex('4c4c')
    assert get_pushdata_code(0xff) == bytes.fromhex('4cff')
    assert get_pushdata_code(0x0100) == bytes.fromhex('4d0001')
    assert get_pushdata_code(0xffff) == bytes.fromhex('4dffff')
    assert get_pushdata_code(0x010000) == bytes.fromhex('4e00000100')
    assert get_pushdata_code(0x01020304) == bytes.fromhex('4e04030201')

    with pytest.raises(ValueError, match=r'data too long to encode in a PUSHDATA opcode'):
        get_pushdata_code(0x0100000000)


def test_encode_pushdata():
    # minimal push
    assert encode_pushdata(b'') == OpCode.OP_0
    assert encode_pushdata(b'\x00') == b'\x01\x00'
    assert encode_pushdata(b'\x01') == OpCode.OP_1
    assert encode_pushdata(b'\x02') == OpCode.OP_2
    assert encode_pushdata(b'\x10') == OpCode.OP_16
    assert encode_pushdata(b'\x11') == b'\x01\x11'
    assert encode_pushdata(b'\x81') == OpCode.OP_1NEGATE
    # non-minimal push
    with pytest.raises(AssertionError, match=r'empty pushdata'):
        encode_pushdata(b'', False)
    assert encode_pushdata(b'\x00', False) == b'\x01\x00'
    assert encode_pushdata(b'\x01', False) == b'\x01\x01'
    assert encode_pushdata(b'\x02', False) == b'\x01\x02'
    assert encode_pushdata(b'\x10', False) == b'\x01\x10'
    assert encode_pushdata(b'\x11', False) == b'\x01\x11'
    assert encode_pushdata(b'\x81', False) == b'\x01\x81'


def test_encode_int():
    assert encode_int(-2147483648) == bytes.fromhex('05 00 00 00 80 80')
    assert encode_int(-2147483647) == bytes.fromhex('04 FF FF FF FF')
    assert encode_int(-8388608) == bytes.fromhex('04 00 00 80 80')
    assert encode_int(-8388607) == bytes.fromhex('03 FF FF FF')
    assert encode_int(-32768) == bytes.fromhex('03 00 80 80')
    assert encode_int(-32767) == bytes.fromhex('02 FF FF')
    assert encode_int(-128) == bytes.fromhex('02 80 80')
    assert encode_int(-127) == bytes.fromhex('01 FF')
    assert encode_int(-17) == bytes.fromhex('01 91')
    assert encode_int(-16) == bytes.fromhex('01 90')
    assert encode_int(-2) == bytes.fromhex('01 82')
    assert encode_int(-1) == OpCode.OP_1NEGATE

    assert encode_int(0) == OpCode.OP_0

    assert encode_int(1) == OpCode.OP_1
    assert encode_int(2) == OpCode.OP_2
    assert encode_int(16) == OpCode.OP_16
    assert encode_int(17) == bytes.fromhex('01 11')
    assert encode_int(127) == bytes.fromhex('01 7F')
    assert encode_int(128) == bytes.fromhex('02 80 00')
    assert encode_int(32767) == bytes.fromhex('02 FF 7F')
    assert encode_int(32768) == bytes.fromhex('03 00 80 00')
    assert encode_int(8388607) == bytes.fromhex('03 FF FF 7F')
    assert encode_int(8388608) == bytes.fromhex('04 00 00 80 00')
    assert encode_int(2147483647) == bytes.fromhex('04 FF FF FF 7F')
    assert encode_int(2147483648) == bytes.fromhex('05 00 00 00 80 00')



import pytest

from bsv.hd.bip32 import Xpub, Xprv, ckd, master_xprv_from_seed
from bsv.hd.bip39 import WordList, mnemonic_from_entropy, seed_from_mnemonic, validate_mnemonic
from bsv.hd.bip44 import derive_xprvs_from_mnemonic, derive_xkeys_from_xkey

_mnemonic = 'slice simple ring fluid capital exhaust will illegal march annual shift hood'
_seed = '4fc3bea5ae2df6c5a93602e87085de5a7c1e94bb7ab5e6122364753cc51aa5e210c32aec1c58ed570c83084ec3b60b4ad69075bc62c05edb8e538ae2843f4f59'

master_xprv = 'xprv9s21ZrQH143K4SSfHuCgyJKsown12SFNpzCf3XYJT67mkaVaWCCBqiGBRZRmgk2ypzXoWzAccyVPGBW69A6LLRMnbY6GZ27q6UkiJDnPjhT'
master_xpub = 'xpub661MyMwAqRbcGvX8PvjhLSGcMycVRtyECD8Fquwv1RekdNpj3jWSPWafGsdNa6TNVmDN9HpPe2tRPofzHTYAUeQFUsAQpzuVSDDyUCt975T'

# m/0
normal_xprv = 'xprv9v35D6cvdU6R1d3UuY6bbR87h6pJLQn3kXY9jwGXhqTX129XT5jZnEyTDoDKnoE9k7HSK7MNv7E3gEGkt4Bp7BkcgHgXUHzQHXueD1t2vRj'
normal_xpub = 'xpub692Rcc9pTqeiE77x1ZdbxZ4rF8enjsVu7kTkYKg9GAzVspUfzd3pL3Hw56Fkgg4vrhayKd6k33uiJgmicfiKf2T1E5brXQLeQni1ake7uSv'

# m/0'
hardened_xprv = 'xprv9v35D6d4y8dP9r1N2koQ49hwzk8EDT4msMFAXGertWPxQDByPqZ1e3k6U34kwU4iCnur3UcxX4SvaDFcrubYd3ktsfpCraGmWpqDq4fm1SJ'
hardened_xpub = 'xpub692Rcc9xoWBgNL5q8nLQRHegYmxicundEaAmKf4USqvwH1X7wNsGBr4aKHLeKDA5ghqECjBErUwLaYZ6As5PpqsFJbZD3jyBWrk6QKG8QQX'


def test_xkey():
    with pytest.raises(TypeError, match=r'unsupported extended key type'):
        # noinspection PyTypeChecker
        Xpub(1)

    assert Xpub.from_xprv(master_xprv) == Xpub(master_xpub)
    assert Xpub.from_xprv(normal_xprv) == Xpub(normal_xpub)
    assert Xpub.from_xprv(Xprv(hardened_xprv)) == Xpub(hardened_xpub)

    assert Xpub(master_xpub).chain_code == Xprv(master_xprv).chain_code

    assert str(Xprv(master_xprv)) == master_xprv
    assert str(Xpub(master_xpub)) == master_xpub

    assert str(Xprv(master_xprv).ckd(0)) == normal_xprv
    assert str(Xprv(master_xprv).ckd('80000000')) == hardened_xprv
    assert str(Xprv(master_xprv).ckd(b'\x80\x00\x00\x00')) == hardened_xprv

    assert str(Xpub(master_xpub).ckd(0)) == normal_xpub
    assert str(Xpub(master_xpub).ckd('00000000')) == normal_xpub
    assert str(Xpub(master_xpub).ckd(b'\x00\x00\x00\x00')) == normal_xpub

    wif = 'KxegHzrskmyDrSuymrQVEWbLjQRm5y7c9XJYoVFAtfi1uszycQX7'
    public_key_hex = '033394416f0d04d0758e002f6708dd121a4c02eae4fee8734fc359c27bd22a92bd'
    address = '1LRax3BdP3SaSnGoD2pkAMTrbuATtog7Kj'
    assert Xprv(normal_xprv).xpub() == Xpub(normal_xpub)
    assert Xprv(normal_xprv).public_key().hex() == public_key_hex
    assert Xprv(normal_xprv).address() == address
    assert Xprv(normal_xprv).private_key().wif() == wif
    assert Xpub(normal_xpub).public_key().hex() == public_key_hex
    assert Xpub(normal_xpub).address() == address

    assert Xprv.from_seed(_seed) == Xprv(master_xprv)
    assert Xprv.from_seed(bytes.fromhex(_seed)) == Xprv(master_xprv)

    assert str(master_xprv_from_seed(_seed)) == master_xprv


def test_ckd():
    assert ckd(Xprv(master_xprv), "m") == Xprv(master_xprv)
    assert ckd(Xprv(master_xprv), ".") == Xprv(master_xprv)
    assert ckd(Xprv(master_xprv), "m/0'") == Xprv(hardened_xprv)
    assert ckd(Xprv(master_xprv), "./0'") == Xprv(hardened_xprv)
    assert ckd(Xpub(master_xpub), 'm/0') == Xpub(normal_xpub)
    assert ckd(Xpub(master_xpub), './0') == Xpub(normal_xpub)

    with pytest.raises(AssertionError, match=r'absolute path for non-master key'):
        ckd(Xpub(normal_xpub), 'm/0')

    with pytest.raises(AssertionError, match=r"can't make hardened derivation from xpub"):
        ckd(Xpub(master_xpub), "m/0'")


def test_wordlist():
    assert WordList.get_word(0) == 'abandon'
    assert WordList.get_word(9) == 'abuse'
    assert WordList.get_word(b'\x01\x02') == 'cake'
    assert WordList.get_word(2047) == 'zoo'
    with pytest.raises(AssertionError, match=r'index out of range'):
        WordList.get_word(2048)
    with pytest.raises(AssertionError, match=r'wordlist not supported'):
        WordList.get_word(0, 'zh-tw')

    assert WordList.index_word('abandon') == 0
    assert WordList.index_word('zoo') == 2047
    with pytest.raises(ValueError, match=r'invalid word'):
        WordList.index_word('hi')


def test_mnemonic():
    assert seed_from_mnemonic(_mnemonic).hex() == _seed

    assert len(mnemonic_from_entropy().split(' ')) == 12

    entropy = '27c715c6caf5b38172ef2b35d51764d5'
    mnemonic = 'chief december immune nominee forest scheme slight tornado cupboard post summer program'
    sd = 'ccf9ff0d7541429ccff7c3c5a03bedd8e736542346f2e020c2151df5169bd14482c761e2cafc9e25990c584867e8b2f2d84ade643109da5e60f1bf03a63c41a7'
    assert mnemonic_from_entropy(entropy) == mnemonic
    assert mnemonic_from_entropy(bytes.fromhex(entropy)) == mnemonic
    assert seed_from_mnemonic(mnemonic).hex() == sd

    entropy = '13b8924d0e0436a6d12200bee8a599c38e31c17ea96a7b58d41b5d3a1aed2339'
    mnemonic = 'beauty setup nation bright drop fat duty divorce same early grid mandate ' \
               'toast thing wide coil kitten shop almost risk payment isolate mind dinner'
    sd = '0c15a3c37a38157147b03225478cdb244b4de24c8da7bd0ccf75893223454caacebae97b5e1d3e966f9a9ce1526944b2b7ca17e21651a0e6f101b01f951008e2'
    assert mnemonic_from_entropy(entropy) == mnemonic
    assert seed_from_mnemonic(mnemonic).hex() == sd

    mnemonic = 'furnace tunnel buyer merry feature stamp brown client fine stomach company blossom'
    sd1 = '2588c36c5d2685b89e5ab06406cd5e96efcc3dc101c4ebd391fc93367e5525aca6c7a5fe4ea8b973c58279be362dbee9a84771707fc6521c374eb10af1044283'
    sd2 = '1e8340ad778a2bbb1ccac4dd02e6985c888a0db0c40d9817998c0ef3da36e846b270f2c51ad67ac6f51183f567fd97c58a31d363296d5dc6245a0a3c4a3e83c5'
    assert seed_from_mnemonic(mnemonic).hex() == sd1
    assert seed_from_mnemonic(mnemonic, passphrase='bitcoin').hex() == sd2

    with pytest.raises(AssertionError, match=r'invalid mnemonic, bad entropy bit length'):
        validate_mnemonic('license expire dragon express pulse behave sibling draft vessel')
    with pytest.raises(AssertionError, match=r'invalid mnemonic, checksum mismatch'):
        validate_mnemonic('dignity candy ostrich wide enrich bubble solid sun cannon deposit merge replace')

    path = "m/44'/0'/0'/0/0"
    mnemonic = '塔 恨 非 送 惨 右 娘 适 呵 二 溶 座 伸 徐 鼓'
    sd = 'fb520b58b6db65172fb00322826a902463b0e6af6f2dfd400ce77b528e81f6cbc785835e7e7f7aec5368916b96607f2a1b348bfa483bf8d3a23acf744b4ce209'
    assert seed_from_mnemonic(mnemonic, lang='zh-cn').hex() == sd
    assert ckd(master_xprv_from_seed(seed_from_mnemonic(mnemonic, 'zh-cn')),
               path).address() == '1C5XJhzRNDDuPNzETmJFFhkU46s1bBFqyV'

    mnemonic = '猛 念 回 风 自 将 大 鸟 说 揭 召 必 旱 济 挡 陆 染 昏'
    sd = '1a9553b9a7d7a394841ca8f5883bf5366c4c7a8ace58b5d32bd291dd9bfa25072253e9904e943ffe426f334bd8275595a87c425f8713b619945155fd5e88a390'
    assert seed_from_mnemonic(mnemonic, lang='zh-cn').hex() == sd
    assert ckd(master_xprv_from_seed(seed_from_mnemonic(mnemonic, 'zh-cn')),
               path).address() == '1GeiN188BR499mp4JvT1EHD7MVUZ1jJVMj'

    mnemonic = '部 街 缓 弯 醒 巧 传 文 馆 央 怕 纬 疾 沸 静 丘 促 罗 辅 追 勃'
    sd = 'cd552980402550f9ec350cd63cb582d1087c333dbf5044c48ee0ec9f083636193b3738ae04d18198476904fdcd5955764b5f5630b0db0d35d311d0a0fd9b7e8d'
    assert seed_from_mnemonic(mnemonic, lang='zh-cn').hex() == sd
    assert ckd(master_xprv_from_seed(seed_from_mnemonic(mnemonic, 'zh-cn')),
               path).address() == '1PUaGha3pSPUwCT7JTLTXUdnL9wbvibU1u'


def test_derive():
    mnemonic = 'chief december immune nominee forest scheme slight tornado cupboard post summer program'

    assert [xprv.private_key().wif() for xprv in derive_xprvs_from_mnemonic(mnemonic, 2, 0, path="m/44'/0'/0'")] == []

    assert [xprv.private_key().wif() for xprv in derive_xprvs_from_mnemonic(mnemonic, 0, 2, path="m/44'/0'/0'")] == [
        'KwW635XeepCG6SzpSMugJ2XDckdnoP6DsDSvg1kjLt11tEJyYaSH',
        'L1QcQMMtXar4nb9hkWdmawumopgKZfRi4Ge1T143w3mBWw7QmuU1',
    ]

    assert [xprv.private_key().wif() for xprv in
            derive_xprvs_from_mnemonic(mnemonic, "1'", "3'", path="m/44'/0'/0'")] == [
               'L3hELjh4wmLgrWEqK2mLsMW3WL3BiYYN3e7wP4s8Xtqi9M8sfNwq',
               'L2orKKStKu1zB2gUzwvEosy8nzohBKBYHZpPThHJ9a6imJs687RA',
           ]

    assert [xprv.private_key().wif() for xprv in
            derive_xprvs_from_mnemonic(mnemonic, 0, 2, change=1, path="m/44'/0'/0'")] == [
               'L4ihevFGHEu3Hdk8TDCucLkyrDSntxhiEnjp2SQARPEnmHXsMG2L',
               'KzRrUofZDgfArmmhqtuS7EMvTUmvWT7BGpqJdCJzmBiwWixatiEk',
           ]

    assert [xprv.private_key().wif() for xprv in
            derive_xprvs_from_mnemonic(mnemonic, 0, 2, change="0'", path="m/44'/0'/0'")] == [
               'L4gRZpDf5Nm6JrowpcX9Z8zmxKNNgiWE61uBb4xF2i8Y9DjXiK5u',
               'KwxW8VrNkoxjjyH22cMPv6ZbBKZKTcV6iSqjTP73daih4fyg3znY',
           ]

    assert [xprv.private_key().wif() for xprv in derive_xprvs_from_mnemonic(mnemonic, 0, 2, path="m/44'/236'/0'")] == [
        'L4toENSefoBpDJcfGAwrSMcyqBNmfSYjgkAP2qeNujw5oPQGvNtM',
        'KzwYj8kMuNqmxLModB1nyPoZjPskCqPXJHf6oUdpHkBK6ZgDUoHE',
    ]

    assert [xprv.private_key().wif() for xprv in
            derive_xprvs_from_mnemonic(mnemonic, 0, 2, passphrase='bitcoin', path="m/44'/0'/0'")] == [
               'L3BWttJh9azQPvvYwFHeEyPniDTCA9TSaPqHKA7jadLVUHDg8KKC',
               'L3h1AvgvscQ1twBTgrH522yNtBfvPjSue3zfH5YRQCt6PdV7FdwS',
           ]

    mnemonic = '安 效 架 碱 皮 伐 鸭 膨 何 泰 陕 森'

    assert [xprv.private_key().wif() for xprv in
            derive_xprvs_from_mnemonic(mnemonic, 0, 2, lang='zh-cn', path="m/44'/0'/0'")] == [
               'KxmA3w8DSR37eD5RqqgkrHHjLgWkZbhyotDd3EehXjvKKziucpwd',
               'L4Q21pxZZpMHWnH19FypFmQhkkxgj1ZSMeCbSfdELu5HnZZm1yJk',
           ]

    xpub = Xpub(
        'xpub6Cz7kFTJ71HQPZpSb8SF2naobZ6HnLgZ8izFEJ31A5R4aR4c3sgHGP8KFwSJbUKLuBeNM4CdXHdrWTqC4sViEHTdv9mXAdCy2E3e6kjUWfB')

    assert [xpub.address() for xpub in derive_xkeys_from_xkey(xpub, 0, 1)] == ['1NDA9czdzkaJFA5Cj1TRyKeews5GrJ9QKR']

    with pytest.raises(AssertionError, match=r"can't make hardened derivation from xpub"):
        derive_xkeys_from_xkey(xpub, "0'", "1'")





from typing import Union, List
from .bip32 import Xprv, Xpub, step_to_index, bip32_derive_xprv_from_mnemonic, bip32_derive_xkeys_from_xkey
from ..constants import Network, BIP44_DERIVATION_PATH


def bip44_derive_xprv_from_mnemonic(mnemonic: str,
                                    lang: str = 'en',
                                    passphrase: str = '',
                                    prefix: str = 'mnemonic',
                                    path: str = BIP44_DERIVATION_PATH,
                                    network: Network = Network.MAINNET) -> Xprv:
    """
    Derives extended private key using BIP44 format- it is a subset of BIP32.
    Inherits from BIP32, only changing the default path value.
    """
    return bip32_derive_xprv_from_mnemonic(
        mnemonic=mnemonic,
        lang=lang,
        passphrase=passphrase,
        prefix=prefix,
        path=path,
        network=network
    )


def bip44_derive_xprvs_from_mnemonic(mnemonic: str,
                                     index_start: Union[str, int],
                                     index_end: Union[str, int],
                                     lang: str = 'en',
                                     passphrase: str = '',
                                     prefix: str = 'mnemonic',
                                     path: str = BIP44_DERIVATION_PATH,
                                     change: Union[str, int] = 0,
                                     network: Network = Network.MAINNET) -> List[Xprv]:
    """
    Derive a range of extended keys from a nmemonic using BIP44 format
    """

    xprv = bip44_derive_xprv_from_mnemonic(mnemonic, lang, passphrase, prefix, path, network)
    return _derive_xkeys_from_xkey(xprv, index_start, index_end, change)


def _derive_xkeys_from_xkey(xkey: Union[Xprv, Xpub],
                            index_start: Union[str, int],
                            index_end: Union[str, int],
                            change: Union[str, int] = 0) -> List[Union[Xprv, Xpub]]:
    """
    this function is internal use only within bip44 module
    """
    change_xkey = xkey.ckd(step_to_index(change))
    return [change_xkey.ckd(i) for i in range(step_to_index(index_start), step_to_index(index_end))]


# [DEPRECATED]
def derive_xkeys_from_xkey(xkey: Union[Xprv, Xpub],
                           index_start: Union[str, int],
                           index_end: Union[str, int],
                           change: Union[str, int] = 0) -> List[Union[Xprv, Xpub]]:
    """
     [DEPRECATED] Use bip32_derive_xkeys_from_xkey instead.
       This function name is kept for backward compatibility.
    """
    return _derive_xkeys_from_xkey(xkey=xkey,
                                   index_start=index_start,
                                   index_end=index_end,
                                   change=change)


# [DEPRECATED]
def derive_xprv_from_mnemonic(mnemonic: str,
                              lang: str = 'en',
                              passphrase: str = '',
                              prefix: str = 'mnemonic',
                              path: str = BIP44_DERIVATION_PATH,
                              network: Network = Network.MAINNET) -> Xprv:
    """
     [DEPRECATED] Use bip44_derive_xprv_from_mnemonic instead.
       This function name is kept for backward compatibility.
    """
    return bip44_derive_xprv_from_mnemonic(
        mnemonic=mnemonic,
        lang=lang,
        passphrase=passphrase,
        prefix=prefix,
        path=path,
        network=network
    )


# [DEPRECATED]
def derive_xprvs_from_mnemonic(mnemonic: str,
                               index_start: Union[str, int],
                               index_end: Union[str, int],
                               lang: str = 'en',
                               passphrase: str = '',
                               prefix: str = 'mnemonic',
                               path: str = BIP44_DERIVATION_PATH,
                               change: Union[str, int] = 0,
                               network: Network = Network.MAINNET) -> List[Xprv]:
    """
     [DEPRECATED] Use bip44_derive_xprvs_from_mnemonic instead.
       This function name is kept for backward compatibility.
    """
    return bip44_derive_xprvs_from_mnemonic(mnemonic=mnemonic,
                                            index_start=index_start,
                                            index_end=index_end,
                                            lang=lang,
                                            passphrase=passphrase,
                                            prefix=prefix,
                                            path=path,
                                            change=change,
                                            network=network)



import pytest 

from bsv.constants import SIGHASH
from bsv.hash import hash256
from bsv.keys import PrivateKey
from bsv.script.script import Script
from bsv.script.type import P2PKH, OpReturn
from bsv.transaction import TransactionInput, TransactionOutput, Transaction
from bsv.transaction_preimage import _preimage, tx_preimages
from bsv.utils import encode_pushdata, Reader
from bsv.fee_models import SatoshisPerKilobyte

digest1 = bytes.fromhex(
    "01000000"
    "ae4b0ed7fb33ec9d5c567520f8cf5f688207f28d5c2f2225c5fe62f7f17c0a25"
    "3bb13029ce7b1f559ef5e747fcac439f1455a2ec7c5f09b72290795e70665044"
    "48dd1f8e77b4a6a75e9b0d0908b25f56b8c98ce37d1fb5ada534d49d0957bcd201000000"
    "1976a9146a176cd51593e00542b8e1958b7da2be97452d0588ac"
    "e803000000000000"
    "ffffffff"
    "048129b26f1d89828c88cdcd472f8f20927822ab7a3d6532cb921c4019f51301"
    "00000000"
    "41000000"
)
digest2 = bytes.fromhex(
    "01000000"
    "ee2851915c957b7187967dabb54f32c00964c689285d3b73e7b2b92e30723c88"
    "752adad0a7b9ceca853768aebb6965eca126a62965f698a0c1bc43d83db632ad"
    "48dd1f8e77b4a6a75e9b0d0908b25f56b8c98ce37d1fb5ada534d49d0957bcd202000000"
    "1976a9146a176cd51593e00542b8e1958b7da2be97452d0588ace"
    "803000000000000"
    "ffffffff"
    "d67a44dde8ee744b7d73b50a3b3a887cb3321d6e16025273f760046c35a265fd"
    "00000000"
    "41000000"
)
digest3 = bytes.fromhex(
    "01000000"
    "ee2851915c957b7187967dabb54f32c00964c689285d3b73e7b2b92e30723c88"
    "752adad0a7b9ceca853768aebb6965eca126a62965f698a0c1bc43d83db632ad"
    "e4c1a33b3a7ca18ef1d6030c6ec222902195f186cb864e09bc1db08b3ea5c1fc00000000"
    "1976a9146a176cd51593e00542b8e1958b7da2be97452d0588ace"
    "803000000000000"
    "ffffffff"
    "d67a44dde8ee744b7d73b50a3b3a887cb3321d6e16025273f760046c35a265fd"
    "00000000"
    "41000000"
)

BRC62Hex = "0100beef01fe636d0c0007021400fe507c0c7aa754cef1f7889d5fd395cf1f785dd7de98eed895dbedfe4e5bc70d1502ac4e164f5bc16746bb0868404292ac8318bbac3800e4aad13a014da427adce3e010b00bc4ff395efd11719b277694cface5aa50d085a0bb81f613f70313acd28cf4557010400574b2d9142b8d28b61d88e3b2c3f44d858411356b49a28a4643b6d1a6a092a5201030051a05fc84d531b5d250c23f4f886f6812f9fe3f402d61607f977b4ecd2701c19010000fd781529d58fc2523cf396a7f25440b409857e7e221766c57214b1d38c7b481f01010062f542f45ea3660f86c013ced80534cb5fd4c19d66c56e7e8c5d4bf2d40acc5e010100b121e91836fd7cd5102b654e9f72f3cf6fdbfd0b161c53a9c54b12c841126331020100000001cd4e4cac3c7b56920d1e7655e7e260d31f29d9a388d04910f1bbd72304a79029010000006b483045022100e75279a205a547c445719420aa3138bf14743e3f42618e5f86a19bde14bb95f7022064777d34776b05d816daf1699493fcdf2ef5a5ab1ad710d9c97bfb5b8f7cef3641210263e2dee22b1ddc5e11f6fab8bcd2378bdd19580d640501ea956ec0e786f93e76ffffffff013e660000000000001976a9146bfd5c7fbe21529d45803dbcf0c87dd3c71efbc288ac0000000001000100000001ac4e164f5bc16746bb0868404292ac8318bbac3800e4aad13a014da427adce3e000000006a47304402203a61a2e931612b4bda08d541cfb980885173b8dcf64a3471238ae7abcd368d6402204cbf24f04b9aa2256d8901f0ed97866603d2be8324c2bfb7a37bf8fc90edd5b441210263e2dee22b1ddc5e11f6fab8bcd2378bdd19580d640501ea956ec0e786f93e76ffffffff013c660000000000001976a9146bfd5c7fbe21529d45803dbcf0c87dd3c71efbc288ac0000000000"
MerkleRootFromBEEF = "bb6f640cc4ee56bf38eb5a1969ac0c16caa2d3d202b22bf3735d10eec0ca6e00"

tx_in = TransactionInput(unlocking_script=Script("ae"))

tx_out = TransactionOutput(locking_script=Script("ae"), satoshis=5)

tx = Transaction(
    tx_inputs=[tx_in],
    tx_outputs=[tx_out],
)
txhex = "000000000100000000000000000000000000000000000000000000000000000000000000000000000001ae0000000001050000000000000001ae00000000"
txbuf = bytes.fromhex(txhex)

tx2idhex = "8c9aa966d35bfeaf031409e0001b90ccdafd8d859799eb945a3c515b8260bcf2"
tx2hex = "01000000029e8d016a7b0dc49a325922d05da1f916d1e4d4f0cb840c9727f3d22ce8d1363f000000008c493046022100e9318720bee5425378b4763b0427158b1051eec8b08442ce3fbfbf7b30202a44022100d4172239ebd701dae2fbaaccd9f038e7ca166707333427e3fb2a2865b19a7f27014104510c67f46d2cbb29476d1f0b794be4cb549ea59ab9cc1e731969a7bf5be95f7ad5e7f904e5ccf50a9dc1714df00fbeb794aa27aaff33260c1032d931a75c56f2ffffffffa3195e7a1ab665473ff717814f6881485dc8759bebe97e31c301ffe7933a656f020000008b48304502201c282f35f3e02a1f32d2089265ad4b561f07ea3c288169dedcf2f785e6065efa022100e8db18aadacb382eed13ee04708f00ba0a9c40e3b21cf91da8859d0f7d99e0c50141042b409e1ebbb43875be5edde9c452c82c01e3903d38fa4fd89f3887a52cb8aea9dc8aec7e2c9d5b3609c03eb16259a2537135a1bf0f9c5fbbcbdbaf83ba402442ffffffff02206b1000000000001976a91420bb5c3bfaef0231dc05190e7f1c8e22e098991e88acf0ca0100000000001976a9149e3e2d23973a04ec1b02be97c30ab9f2f27c3b2c88ac00000000"
tx2buf = bytes.fromhex(tx2hex)


def test_new_tx():
    tx = Transaction()

    assert Transaction.from_hex(txbuf).hex() == txhex

    # should set known defaults
    assert tx.version == 1
    assert len(tx.inputs) == 0
    assert len(tx.outputs) == 0
    assert tx.locktime == 0


def test_transaction_from_hex():
    assert Transaction.from_hex(txhex).hex() == txhex
    assert Transaction.from_hex(tx2hex).hex() == tx2hex


def test_transaction_parse_script_offsets():
    tx = Transaction.from_hex(tx2buf)
    assert tx.txid() == tx2idhex
    r = Transaction.parse_script_offsets(tx2buf)
    assert len(r["inputs"]) == 2
    assert len(r["outputs"]) == 2
    for vin in range(2):
        i = r["inputs"][vin]
        script = tx2buf[i["offset"] : i["offset"] + i["length"]]
        assert script == tx.inputs[vin].unlocking_script.serialize()
    for vout in range(2):
        o = r["outputs"][vout]
        script = tx2buf[o["offset"] : o["offset"] + o["length"]]
        assert script == tx.outputs[vout].locking_script.serialize()


def test_transaction_to_hex():
    assert Transaction.from_hex(txhex).hex() == txhex


def test_transaction_serialize():
    assert Transaction.from_hex(txbuf).serialize().hex() == txhex


def test_transaction_hash():
    tx = Transaction.from_hex(tx2buf)
    assert tx.hash()[::-1].hex() == tx2idhex


def test_transaction_id():
    tx = Transaction.from_hex(tx2buf)
    assert tx.txid() == tx2idhex


def test_transaction_add_input():
    tx_in = TransactionInput()
    tx = Transaction()
    assert len(tx.inputs) == 0
    tx.add_input(tx_in)
    assert len(tx.inputs) == 1


def test_transaction_add_output():
    tx_out = TransactionOutput(locking_script=Script("6a"), satoshis=0)
    tx = Transaction()
    assert len(tx.outputs) == 0
    tx.add_output(tx_out)
    assert len(tx.outputs) == 1


def test_transaction_signing_hydrate_scripts():
    private_key = PrivateKey(
        bytes.fromhex(
            "f97c89aaacf0cd2e47ddbacc97dae1f88bec49106ac37716c451dcdd008a4b62"
        )
    )
    public_key = private_key.public_key()
    public_key_hash = public_key.address()

    source_tx = Transaction(
        [], [TransactionOutput(P2PKH().lock(public_key_hash), 4000)]
    )
    spend_tx = Transaction(
        [
            TransactionInput(
                source_transaction=source_tx,
                source_output_index=0,
                unlocking_script_template=P2PKH().unlock(private_key),
            )
        ],
        [
            TransactionOutput(
                P2PKH().lock(public_key_hash),
                1000,
            ),
            TransactionOutput(
                P2PKH().lock(public_key_hash),
                change=True,
            ),
        ],
    )

    assert not spend_tx.inputs[0].unlocking_script

    spend_tx.fee()
    spend_tx.sign()
    assert spend_tx.inputs[0].unlocking_script


def test_estimated_byte_length():
    _in = TransactionInput(
        source_txid="00" * 32,
        unlocking_script=None,
        unlocking_script_template=P2PKH().unlock(PrivateKey()),
    )
    _in.satoshis = 2000

    _out = TransactionOutput(P2PKH().lock(PrivateKey().address()), 1000)

    t = Transaction().add_input(_in).add_output(_out)

    _in.private_keys = [PrivateKey()]
    assert t.estimated_byte_length() == 192

    _in.unlocking_script = b""
    assert t.estimated_byte_length() == 85
    assert t.estimated_byte_length() == t.byte_length()


def test_beef_serialization():
    brc62_hex = "0100beef01fe636d0c0007021400fe507c0c7aa754cef1f7889d5fd395cf1f785dd7de98eed895dbedfe4e5bc70d1502ac4e164f5bc16746bb0868404292ac8318bbac3800e4aad13a014da427adce3e010b00bc4ff395efd11719b277694cface5aa50d085a0bb81f613f70313acd28cf4557010400574b2d9142b8d28b61d88e3b2c3f44d858411356b49a28a4643b6d1a6a092a5201030051a05fc84d531b5d250c23f4f886f6812f9fe3f402d61607f977b4ecd2701c19010000fd781529d58fc2523cf396a7f25440b409857e7e221766c57214b1d38c7b481f01010062f542f45ea3660f86c013ced80534cb5fd4c19d66c56e7e8c5d4bf2d40acc5e010100b121e91836fd7cd5102b654e9f72f3cf6fdbfd0b161c53a9c54b12c841126331020100000001cd4e4cac3c7b56920d1e7655e7e260d31f29d9a388d04910f1bbd72304a79029010000006b483045022100e75279a205a547c445719420aa3138bf14743e3f42618e5f86a19bde14bb95f7022064777d34776b05d816daf1699493fcdf2ef5a5ab1ad710d9c97bfb5b8f7cef3641210263e2dee22b1ddc5e11f6fab8bcd2378bdd19580d640501ea956ec0e786f93e76ffffffff013e660000000000001976a9146bfd5c7fbe21529d45803dbcf0c87dd3c71efbc288ac0000000001000100000001ac4e164f5bc16746bb0868404292ac8318bbac3800e4aad13a014da427adce3e000000006a47304402203a61a2e931612b4bda08d541cfb980885173b8dcf64a3471238ae7abcd368d6402204cbf24f04b9aa2256d8901f0ed97866603d2be8324c2bfb7a37bf8fc90edd5b441210263e2dee22b1ddc5e11f6fab8bcd2378bdd19580d640501ea956ec0e786f93e76ffffffff013c660000000000001976a9146bfd5c7fbe21529d45803dbcf0c87dd3c71efbc288ac0000000000"
    t = Transaction.from_beef(bytes.fromhex(brc62_hex))
    assert t.inputs[0].source_transaction.merkle_path.block_height == 814435
    beef = t.to_beef()
    assert beef.hex() == brc62_hex


def test_from_reader():
    assert TransactionInput.from_hex("") is None
    tx_in = TransactionInput.from_hex("0011" * 16 + "00112233" + "00" + "00112233")
    assert tx_in.source_txid == "1100" * 16
    assert tx_in.source_output_index == 0x33221100
    assert tx_in.unlocking_script == Script()
    assert tx_in.sequence == 0x33221100

    assert TransactionOutput.from_hex("") is None
    assert Transaction.from_hex("") is None

    t_hex = (
        "01000000"
        + "03"
        + "7a7b64d59a072867d7453b2eb67e0fb883af0f435cbbeffc2bb5a4b13e3f6e08"
        + "01000000"
        + "6b"
        + "483045"
        + "0221008b6f070f73242c7c8c654f493dd441d46dc7b2365c8e9e4c62732da0fb535c58"
        + "02204b96edfb934d08ad0cfaa9bf75887bd8541498fbe19189d45683dcbd0785d0df"
        + "41"
        + "2102e46dcd7991e5a4bd642739249b0158312e1aee56a60fd1bf622172ffe65bd789"
        + "ffffffff"
        + "7a7b64d59a072867d7453b2eb67e0fb883af0f435cbbeffc2bb5a4b13e3f6e08"
        + "03000000"
        + "6a"
        + "473044"
        + "0220501dae7c51c6e5cb0f12a635ccbc61e283cb2e838d624d7df7f1ba1b0ab2087b"
        + "02207f67f3883735464f6067357c901fc1b8ddf8bf8695b54b2790d6a0106acf2340"
        + "41"
        + "2102e46dcd7991e5a4bd642739249b0158312e1aee56a60fd1bf622172ffe65bd789"
        + "ffffffff"
        + "7a7b64d59a072867d7453b2eb67e0fb883af0f435cbbeffc2bb5a4b13e3f6e08"
        + "02000000"
        + "8b"
        + "483045"
        + "022100b04829882018f7488508cb8587612fb017584ffc2b4d22e4300b95178be642a3"
        + "02207937cb643eef061b53704144148bec25645fbbaf4eedd5586ad9b018d4f6c9d441"
        + "41"
        + "04"
        + "e46dcd7991e5a4bd642739249b0158312e1aee56a60fd1bf622172ffe65bd78997693d32c540ac253de7a3dc73f7e4ba7b38d2dc1ecc8e07920b496fb107d6b2"
        + "ffffffff"
        + "02"
        + "0a1a000000000000"
        + "1976a9146a176cd51593e00542b8e1958b7da2be97452d0588ac"
        + "05ea1c0000000000"
        + "1976a9146a176cd51593e00542b8e1958b7da2be97452d0588ac"
        + "00000000"
    )

    r = Reader(bytes.fromhex(t_hex))
    t = Transaction.from_reader(r)
    assert (
        t.txid() == "e8c6b26f26d90e9cf035762a91479635a75eff2b3b2845663ed72a2397acdfd2"
    )


def test_from_hex():
    assert TransactionInput.from_hex("") is None
    tx_in = TransactionInput.from_hex("0011" * 16 + "00112233" + "00" + "00112233")
    assert tx_in.source_txid == "1100" * 16
    assert tx_in.source_output_index == 0x33221100
    assert tx_in.unlocking_script == Script()
    assert tx_in.sequence == 0x33221100

    assert TransactionOutput.from_hex("") is None
    assert Transaction.from_hex("") is None

    t = Transaction.from_hex(
        "01000000"
        + "03"
        + "7a7b64d59a072867d7453b2eb67e0fb883af0f435cbbeffc2bb5a4b13e3f6e08"
        + "01000000"
        + "6b"
        + "483045"
        + "0221008b6f070f73242c7c8c654f493dd441d46dc7b2365c8e9e4c62732da0fb535c58"
        + "02204b96edfb934d08ad0cfaa9bf75887bd8541498fbe19189d45683dcbd0785d0df"
        + "41"
        + "2102e46dcd7991e5a4bd642739249b0158312e1aee56a60fd1bf622172ffe65bd789"
        + "ffffffff"
        + "7a7b64d59a072867d7453b2eb67e0fb883af0f435cbbeffc2bb5a4b13e3f6e08"
        + "03000000"
        + "6a"
        + "473044"
        + "0220501dae7c51c6e5cb0f12a635ccbc61e283cb2e838d624d7df7f1ba1b0ab2087b"
        + "02207f67f3883735464f6067357c901fc1b8ddf8bf8695b54b2790d6a0106acf2340"
        + "41"
        + "2102e46dcd7991e5a4bd642739249b0158312e1aee56a60fd1bf622172ffe65bd789"
        + "ffffffff"
        + "7a7b64d59a072867d7453b2eb67e0fb883af0f435cbbeffc2bb5a4b13e3f6e08"
        + "02000000"
        + "8b"
        + "483045"
        + "022100b04829882018f7488508cb8587612fb017584ffc2b4d22e4300b95178be642a3"
        + "02207937cb643eef061b53704144148bec25645fbbaf4eedd5586ad9b018d4f6c9d441"
        + "41"
        + "04"
        + "e46dcd7991e5a4bd642739249b0158312e1aee56a60fd1bf622172ffe65bd78997693d32c540ac253de7a3dc73f7e4ba7b38d2dc1ecc8e07920b496fb107d6b2"
        + "ffffffff"
        + "02"
        + "0a1a000000000000"
        + "1976a9146a176cd51593e00542b8e1958b7da2be97452d0588ac"
        + "05ea1c0000000000"
        + "1976a9146a176cd51593e00542b8e1958b7da2be97452d0588ac"
        + "00000000"
    )
    assert (
        t.txid() == "e8c6b26f26d90e9cf035762a91479635a75eff2b3b2845663ed72a2397acdfd2"
    )


def test_transaction_bytes_io():
    io = Reader(
        bytes.fromhex(
            "0011223344556677889912fd1234fe12345678ff1234567890abcdef00112233"
        )
    )

    assert io.read_bytes(4) == bytes.fromhex("00112233")
    assert io.read_int(1) == int.from_bytes(bytes.fromhex("44"), "little")
    assert io.read_int(2) == int.from_bytes(bytes.fromhex("5566"), "little")
    assert io.read_int(3, "big") == int.from_bytes(bytes.fromhex("778899"), "big")
    assert io.read_var_int_num() == int.from_bytes(bytes.fromhex("12"), "little")
    assert io.read_var_int_num() == int.from_bytes(bytes.fromhex("1234"), "little")
    assert io.read_var_int_num() == int.from_bytes(bytes.fromhex("12345678"), "little")
    assert io.read_var_int_num() == int.from_bytes(
        bytes.fromhex("1234567890abcdef"), "little"
    )

    assert io.read_bytes(0) == b""
    assert io.read_bytes() == bytes.fromhex("00112233")
    assert io.read_bytes() == b""
    assert io.read_bytes(1) == b""

    assert io.read_int(1) is None
    assert io.read_var_int_num() is None


BRC62Hex = "0100beef01fe636d0c0007021400fe507c0c7aa754cef1f7889d5fd395cf1f785dd7de98eed895dbedfe4e5bc70d1502ac4e164f5bc16746bb0868404292ac8318bbac3800e4aad13a014da427adce3e010b00bc4ff395efd11719b277694cface5aa50d085a0bb81f613f70313acd28cf4557010400574b2d9142b8d28b61d88e3b2c3f44d858411356b49a28a4643b6d1a6a092a5201030051a05fc84d531b5d250c23f4f886f6812f9fe3f402d61607f977b4ecd2701c19010000fd781529d58fc2523cf396a7f25440b409857e7e221766c57214b1d38c7b481f01010062f542f45ea3660f86c013ced80534cb5fd4c19d66c56e7e8c5d4bf2d40acc5e010100b121e91836fd7cd5102b654e9f72f3cf6fdbfd0b161c53a9c54b12c841126331020100000001cd4e4cac3c7b56920d1e7655e7e260d31f29d9a388d04910f1bbd72304a79029010000006b483045022100e75279a205a547c445719420aa3138bf14743e3f42618e5f86a19bde14bb95f7022064777d34776b05d816daf1699493fcdf2ef5a5ab1ad710d9c97bfb5b8f7cef3641210263e2dee22b1ddc5e11f6fab8bcd2378bdd19580d640501ea956ec0e786f93e76ffffffff013e660000000000001976a9146bfd5c7fbe21529d45803dbcf0c87dd3c71efbc288ac0000000001000100000001ac4e164f5bc16746bb0868404292ac8318bbac3800e4aad13a014da427adce3e000000006a47304402203a61a2e931612b4bda08d541cfb980885173b8dcf64a3471238ae7abcd368d6402204cbf24f04b9aa2256d8901f0ed97866603d2be8324c2bfb7a37bf8fc90edd5b441210263e2dee22b1ddc5e11f6fab8bcd2378bdd19580d640501ea956ec0e786f93e76ffffffff013c660000000000001976a9146bfd5c7fbe21529d45803dbcf0c87dd3c71efbc288ac0000000000"


def test_output():
    assert TransactionOutput(
        locking_script=OpReturn().lock(["123", "456"])
    ).locking_script == Script("006a" + "03313233" + "03343536")


def test_digest():
    address = "1AfxgwYJrBgriZDLryfyKuSdBsi59jeBX9"
    # https://whatsonchain.com/tx/4674da699de44c9c5d182870207ba89e5ccf395e5101dab6b0900bbf2f3b16cb
    expected_digest = [digest1]
    t: Transaction = Transaction()
    t_in = TransactionInput(
        source_transaction=Transaction(
            [],
            [
                None,
                TransactionOutput(locking_script=P2PKH().lock(address), satoshis=1000),
            ],
        ),
        source_txid="d2bc57099dd434a5adb51f7de38cc9b8565fb208090d9b5ea7a6b4778e1fdd48",
        source_output_index=1,
        unlocking_script_template=P2PKH().unlock(PrivateKey()),
    )
    t.add_input(t_in)
    t.add_output(
        TransactionOutput(
            locking_script=P2PKH().lock("1JDZRGf5fPjGTpqLNwjHFFZnagcZbwDsxw"),
            satoshis=800,
        )
    )
    assert tx_preimages(t.inputs, t.outputs, t.version, t.locktime) == expected_digest

    # https://whatsonchain.com/tx/c04bbd007ad3987f9b2ea8534175b5e436e43d64471bf32139b5851adf9f477e
    expected_digest = [digest2, digest3]
    t: Transaction = Transaction()
    t_in1 = TransactionInput(
        source_transaction=Transaction(
            [],
            [
                None,
                None,
                TransactionOutput(locking_script=P2PKH().lock(address), satoshis=1000),
            ],
        ),
        source_txid="d2bc57099dd434a5adb51f7de38cc9b8565fb208090d9b5ea7a6b4778e1fdd48",
        source_output_index=2,
        unlocking_script_template=P2PKH().lock(address),
    )
    t_in2 = TransactionInput(
        source_transaction=Transaction(
            [], [TransactionOutput(locking_script=P2PKH().lock(address), satoshis=1000)]
        ),
        source_txid="fcc1a53e8bb01dbc094e86cb86f195219022c26e0c03d6f18ea17c3a3ba3c1e4",
        source_output_index=0,
        unlocking_script_template=P2PKH().unlock(PrivateKey()),
    )
    t.add_inputs([t_in1, t_in2])
    t.add_output(
        TransactionOutput(
            P2PKH().lock("18CgRLx9hFZqDZv75J5kED7ANnDriwvpi1"), satoshis=1700
        )
    )
    assert t.preimage(0) == expected_digest[0]
    assert t.preimage(1) == expected_digest[1]


def test_transaction():
    address = "1AfxgwYJrBgriZDLryfyKuSdBsi59jeBX9"
    t = Transaction()
    t_in = TransactionInput(
        source_transaction=Transaction(
            [],
            [
                None,
                TransactionOutput(locking_script=P2PKH().lock(address), satoshis=1000),
            ],
        ),
        source_txid="d2bc57099dd434a5adb51f7de38cc9b8565fb208090d9b5ea7a6b4778e1fdd48",
        source_output_index=1,
        unlocking_script_template=P2PKH().unlock(PrivateKey()),
    )
    t.add_input(t_in)
    t.add_output(
        TransactionOutput(
            P2PKH().lock("1JDZRGf5fPjGTpqLNwjHFFZnagcZbwDsxw"), satoshis=800
        )
    )

    signature = bytes.fromhex(
        "3044"
        "02207e2c6eb8c4b20e251a71c580373a2836e209c50726e5f8b0f4f59f8af00eee1a"
        "022019ae1690e2eb4455add6ca5b86695d65d3261d914bc1d7abb40b188c7f46c9a5"
    )
    sighash = bytes.fromhex("41")
    public_key = bytes.fromhex(
        "02e46dcd7991e5a4bd642739249b0158312e1aee56a60fd1bf622172ffe65bd789"
    )
    t.inputs[0].unlocking_script = Script(
        encode_pushdata(signature + sighash) + encode_pushdata(public_key)
    )

    assert (
        t.txid() == "4674da699de44c9c5d182870207ba89e5ccf395e5101dab6b0900bbf2f3b16cb"
    )
    assert t.get_fee() == 200
    assert t.byte_length() == 191

    t.inputs[0].sighash = SIGHASH.NONE_ANYONECANPAY_FORKID
    assert t.preimage(0) == _preimage(
        t.inputs[0], t.version, t.locktime, b"\x00" * 32, b"\x00" * 32, b"\x00" * 32
    )
    t.inputs[0].sighash = SIGHASH.SINGLE_ANYONECANPAY_FORKID
    assert t.preimage(0) == _preimage(
        t.inputs[0],
        t.version,
        t.locktime,
        b"\x00" * 32,
        b"\x00" * 32,
        hash256(t.outputs[0].serialize()),
    )

    t.inputs[0].private_keys = [
        PrivateKey("L5agPjZKceSTkhqZF2dmFptT5LFrbr6ZGPvP7u4A6dvhTrr71WZ9")
    ]

    t.outputs[0].satoshis = 100
    t.add_output(TransactionOutput(P2PKH().lock(address), change=True))

    t.fee(SatoshisPerKilobyte(500))

    # 1-2 transaction 226 bytes --> fee 113 satoshi --> 787 left
    assert len(t.outputs) == 2
    assert t.outputs[1].locking_script == P2PKH().lock(address)
    assert t.outputs[1].satoshis == 787


def test_transaction_bytes_io():
    io = Reader(
        bytes.fromhex(
            "0011223344556677889912fd1234fe12345678ff1234567890abcdef00112233"
        )
    )

    assert io.read_bytes(4) == bytes.fromhex("00112233")
    assert io.read_int(1) == int.from_bytes(bytes.fromhex("44"), "little")
    assert io.read_int(2) == int.from_bytes(bytes.fromhex("5566"), "little")
    assert io.read_int(3, "big") == int.from_bytes(bytes.fromhex("778899"), "big")
    assert io.read_var_int_num() == int.from_bytes(bytes.fromhex("12"), "little")
    assert io.read_var_int_num() == int.from_bytes(bytes.fromhex("1234"), "little")
    assert io.read_var_int_num() == int.from_bytes(bytes.fromhex("12345678"), "little")
    assert io.read_var_int_num() == int.from_bytes(
        bytes.fromhex("1234567890abcdef"), "little"
    )

    assert io.read_bytes(0) == b""
    assert io.read_bytes() == bytes.fromhex("00112233")
    assert io.read_bytes() == b""
    assert io.read_bytes(1) == b""

    assert io.read_int(1) is None
    assert io.read_var_int_num() is None


def test_from_hex():
    assert TransactionInput.from_hex("") is None
    tx_in = TransactionInput.from_hex("0011" * 16 + "00112233" + "00" + "00112233")
    assert tx_in.source_txid == "1100" * 16
    assert tx_in.source_output_index == 0x33221100
    assert tx_in.unlocking_script == Script()
    assert tx_in.sequence == 0x33221100

    assert TransactionOutput.from_hex("") is None
    assert Transaction.from_hex("") is None

    t = Transaction.from_hex(
        "01000000"
        + "03"
        + "7a7b64d59a072867d7453b2eb67e0fb883af0f435cbbeffc2bb5a4b13e3f6e08"
        + "01000000"
        + "6b"
        + "483045"
        + "0221008b6f070f73242c7c8c654f493dd441d46dc7b2365c8e9e4c62732da0fb535c58"
        + "02204b96edfb934d08ad0cfaa9bf75887bd8541498fbe19189d45683dcbd0785d0df"
        + "41"
        + "2102e46dcd7991e5a4bd642739249b0158312e1aee56a60fd1bf622172ffe65bd789"
        + "ffffffff"
        + "7a7b64d59a072867d7453b2eb67e0fb883af0f435cbbeffc2bb5a4b13e3f6e08"
        + "03000000"
        + "6a"
        + "473044"
        + "0220501dae7c51c6e5cb0f12a635ccbc61e283cb2e838d624d7df7f1ba1b0ab2087b"
        + "02207f67f3883735464f6067357c901fc1b8ddf8bf8695b54b2790d6a0106acf2340"
        + "41"
        + "2102e46dcd7991e5a4bd642739249b0158312e1aee56a60fd1bf622172ffe65bd789"
        + "ffffffff"
        + "7a7b64d59a072867d7453b2eb67e0fb883af0f435cbbeffc2bb5a4b13e3f6e08"
        + "02000000"
        + "8b"
        + "483045"
        + "022100b04829882018f7488508cb8587612fb017584ffc2b4d22e4300b95178be642a3"
        + "02207937cb643eef061b53704144148bec25645fbbaf4eedd5586ad9b018d4f6c9d441"
        + "41"
        + "04"
        + "e46dcd7991e5a4bd642739249b0158312e1aee56a60fd1bf622172ffe65bd78997693d32c540ac253de7a3dc73f7e4ba7b38d2dc1ecc8e07920b496fb107d6b2"
        + "ffffffff"
        + "02"
        + "0a1a000000000000"
        + "1976a9146a176cd51593e00542b8e1958b7da2be97452d0588ac"
        + "05ea1c0000000000"
        + "1976a9146a176cd51593e00542b8e1958b7da2be97452d0588ac"
        + "00000000"
    )
    assert (
        t.txid() == "e8c6b26f26d90e9cf035762a91479635a75eff2b3b2845663ed72a2397acdfd2"
    )


def test_from_reader():
    assert TransactionInput.from_hex("") is None
    tx_in = TransactionInput.from_hex("0011" * 16 + "00112233" + "00" + "00112233")
    assert tx_in.source_txid == "1100" * 16
    assert tx_in.source_output_index == 0x33221100
    assert tx_in.unlocking_script == Script()
    assert tx_in.sequence == 0x33221100

    assert TransactionOutput.from_hex("") is None
    assert Transaction.from_hex("") is None

    t_hex = (
        "01000000"
        + "03"
        + "7a7b64d59a072867d7453b2eb67e0fb883af0f435cbbeffc2bb5a4b13e3f6e08"
        + "01000000"
        + "6b"
        + "483045"
        + "0221008b6f070f73242c7c8c654f493dd441d46dc7b2365c8e9e4c62732da0fb535c58"
        + "02204b96edfb934d08ad0cfaa9bf75887bd8541498fbe19189d45683dcbd0785d0df"
        + "41"
        + "2102e46dcd7991e5a4bd642739249b0158312e1aee56a60fd1bf622172ffe65bd789"
        + "ffffffff"
        + "7a7b64d59a072867d7453b2eb67e0fb883af0f435cbbeffc2bb5a4b13e3f6e08"
        + "03000000"
        + "6a"
        + "473044"
        + "0220501dae7c51c6e5cb0f12a635ccbc61e283cb2e838d624d7df7f1ba1b0ab2087b"
        + "02207f67f3883735464f6067357c901fc1b8ddf8bf8695b54b2790d6a0106acf2340"
        + "41"
        + "2102e46dcd7991e5a4bd642739249b0158312e1aee56a60fd1bf622172ffe65bd789"
        + "ffffffff"
        + "7a7b64d59a072867d7453b2eb67e0fb883af0f435cbbeffc2bb5a4b13e3f6e08"
        + "02000000"
        + "8b"
        + "483045"
        + "022100b04829882018f7488508cb8587612fb017584ffc2b4d22e4300b95178be642a3"
        + "02207937cb643eef061b53704144148bec25645fbbaf4eedd5586ad9b018d4f6c9d441"
        + "41"
        + "04"
        + "e46dcd7991e5a4bd642739249b0158312e1aee56a60fd1bf622172ffe65bd78997693d32c540ac253de7a3dc73f7e4ba7b38d2dc1ecc8e07920b496fb107d6b2"
        + "ffffffff"
        + "02"
        + "0a1a000000000000"
        + "1976a9146a176cd51593e00542b8e1958b7da2be97452d0588ac"
        + "05ea1c0000000000"
        + "1976a9146a176cd51593e00542b8e1958b7da2be97452d0588ac"
        + "00000000"
    )

    r = Reader(bytes.fromhex(t_hex))
    t = Transaction.from_reader(r)
    assert (
        t.txid() == "e8c6b26f26d90e9cf035762a91479635a75eff2b3b2845663ed72a2397acdfd2"
    )


def test_beef_serialization():
    t = Transaction.from_beef(bytes.fromhex(BRC62Hex))
    assert t.inputs[0].source_transaction.merkle_path.block_height == 814435
    beef = t.to_beef()
    assert beef.hex() == BRC62Hex


def test_ef_serialization():
    tx = Transaction.from_beef(bytes.fromhex(BRC62Hex))
    ef = tx.to_ef()
    expected_ef = "010000000000000000ef01ac4e164f5bc16746bb0868404292ac8318bbac3800e4aad13a014da427adce3e000000006a47304402203a61a2e931612b4bda08d541cfb980885173b8dcf64a3471238ae7abcd368d6402204cbf24f04b9aa2256d8901f0ed97866603d2be8324c2bfb7a37bf8fc90edd5b441210263e2dee22b1ddc5e11f6fab8bcd2378bdd19580d640501ea956ec0e786f93e76ffffffff3e660000000000001976a9146bfd5c7fbe21529d45803dbcf0c87dd3c71efbc288ac013c660000000000001976a9146bfd5c7fbe21529d45803dbcf0c87dd3c71efbc288ac00000000"
    assert ef.hex() == expected_ef

    tx = Transaction.from_hex(
        "0100000001478a4ac0c8e4dae42db983bc720d95ed2099dec4c8c3f2d9eedfbeb74e18cdbb1b0100006b483045022100b05368f9855a28f21d3cb6f3e278752d3c5202f1de927862bbaaf5ef7d67adc50220728d4671cd4c34b1fa28d15d5cd2712b68166ea885522baa35c0b9e399fe9ed74121030d4ad284751daf629af387b1af30e02cf5794139c4e05836b43b1ca376624f7fffffffff01000000000000000070006a0963657274696861736822314c6d763150594d70387339594a556e374d3948565473446b64626155386b514e4a406164386337373536356335363935353261626463636634646362353537376164633936633866613933623332663630373865353664666232326265623766353600000000"
    )

    prev_tx_outs = [None] * 501
    prev_tx_outs[283] = TransactionOutput(
        locking_script=Script("76a9140c77a935b45abdcf3e472606d3bc647c5cc0efee88ac"),
        satoshis=16,
    )
    prev_tx = Transaction([], prev_tx_outs)
    tx.inputs[0].source_transaction = prev_tx

    ef = tx.to_ef()
    expected_ef = "010000000000000000ef01478a4ac0c8e4dae42db983bc720d95ed2099dec4c8c3f2d9eedfbeb74e18cdbb1b0100006b483045022100b05368f9855a28f21d3cb6f3e278752d3c5202f1de927862bbaaf5ef7d67adc50220728d4671cd4c34b1fa28d15d5cd2712b68166ea885522baa35c0b9e399fe9ed74121030d4ad284751daf629af387b1af30e02cf5794139c4e05836b43b1ca376624f7fffffffff10000000000000001976a9140c77a935b45abdcf3e472606d3bc647c5cc0efee88ac01000000000000000070006a0963657274696861736822314c6d763150594d70387339594a556e374d3948565473446b64626155386b514e4a406164386337373536356335363935353261626463636634646362353537376164633936633866613933623332663630373865353664666232326265623766353600000000"
    assert ef.hex() == expected_ef


def test_input_auto_txid():
    prev_tx = Transaction.from_hex('0100000001478a4ac0c8e4dae42db983bc720d95ed2099dec4c8c3f2d9eedfbeb74e18cdbb1b0100006b483045022100b05368f9855a28f21d3cb6f3e278752d3c5202f1de927862bbaaf5ef7d67adc50220728d4671cd4c34b1fa28d15d5cd2712b68166ea885522baa35c0b9e399fe9ed74121030d4ad284751daf629af387b1af30e02cf5794139c4e05836b43b1ca376624f7fffffffff01000000000000000070006a0963657274696861736822314c6d763150594d70387339594a556e374d3948565473446b64626155386b514e4a406164386337373536356335363935353261626463636634646362353537376164633936633866613933623332663630373865353664666232326265623766353600000000')
    
    private_key = PrivateKey("L5agPjZKceSTkhqZF2dmFptT5LFrbr6ZGPvP7u4A6dvhTrr71WZ9")

    tx_in = TransactionInput(
        source_transaction=prev_tx,
        source_output_index=0,
        unlocking_script_template=P2PKH().unlock(private_key),
    )
    
    assert tx_in.source_txid == 'e6adcaf6b86fb5d690a3bade36011cd02f80dd364f1ecf2bb04902aa1b6bf455'
    
    prev_tx.outputs[0].locking_script = None
    with pytest.raises(Exception):
        tx_in = TransactionInput(
            source_transaction=prev_tx,
            source_output_index=0,
            unlocking_script_template=P2PKH().unlock(private_key),
        )


def test_transaction_fee_with_default_rate():
    from bsv.constants import TRANSACTION_FEE_RATE

    address = "1AfxgwYJrBgriZDLryfyKuSdBsi59jeBX9"
    t = Transaction()
    t_in = TransactionInput(
        source_transaction=Transaction(
            [],
            [
                None,
                TransactionOutput(locking_script=P2PKH().lock(address), satoshis=1000),
            ],
        ),
        source_txid="d2bc57099dd434a5adb51f7de38cc9b8565fb208090d9b5ea7a6b4778e1fdd48",
        source_output_index=1,
        unlocking_script_template=P2PKH().unlock(PrivateKey()),
    )
    t.add_input(t_in)
    t.add_output(
        TransactionOutput(
            P2PKH().lock("1JDZRGf5fPjGTpqLNwjHFFZnagcZbwDsxw"), satoshis=100
        )
    )
    t.add_output(TransactionOutput(P2PKH().lock(address), change=True))

    t.fee()

    estimated_size = t.estimated_byte_length()
    expected_fee = int((estimated_size / 1000) * TRANSACTION_FEE_RATE)
    actual_fee = t.get_fee()

    assert abs(actual_fee - expected_fee) <= 1

# TODO: Test tx.verify()



import os
from contextlib import suppress
from hashlib import pbkdf2_hmac
from secrets import randbits
from typing import List, Dict, Union

from ..constants import BIP39_ENTROPY_BIT_LENGTH_LIST, BIP39_ENTROPY_BIT_LENGTH
from ..hash import sha256
from ..utils import bytes_to_bits, bits_to_bytes


class WordList:
    """
    BIP39 word list
    """
    LIST_WORDS_COUNT: int = 2048

    path = os.path.join(os.path.dirname(__file__), 'wordlist')
    #
    # en
    #   https://github.com/bitcoin/bips/blob/master/bip-0039/english.txt
    # zh-cn
    #   https://github.com/bitcoin/bips/blob/master/bip-0039/chinese_simplified.txt
    #
    files: Dict[str, str] = {
        'en': os.path.join(path, 'english.txt'),
        'zh-cn': os.path.join(path, 'chinese_simplified.txt'),
    }
    wordlist: Dict[str, List[str]] = {}

    @classmethod
    def load(cls) -> None:
        for lang in WordList.files.keys():
            if not WordList.wordlist.get(lang):
                WordList.wordlist[lang] = WordList.load_wordlist(lang)

    @classmethod
    def load_wordlist(cls, lang: str = 'en') -> List[str]:
        assert lang in WordList.files.keys(), f'{lang} wordlist not supported'
        with open(WordList.files[lang], 'r', encoding='utf-8') as f:
            words: List[str] = f.read().splitlines()
        assert len(words) == WordList.LIST_WORDS_COUNT, 'broken wordlist file'
        return words

    @classmethod
    def get_word(cls, index: Union[int, bytes], lang: str = 'en') -> str:
        WordList.load()
        assert lang in WordList.wordlist.keys(), f'{lang} wordlist not supported'
        if isinstance(index, bytes):
            index = int.from_bytes(index, 'big')
        assert 0 <= index < WordList.LIST_WORDS_COUNT, 'index out of range'
        return WordList.wordlist[lang][index]

    @classmethod
    def index_word(cls, word: str, lang: str = 'en') -> int:
        WordList.load()
        assert lang in WordList.wordlist.keys(), f'{lang} wordlist not supported'
        with suppress(Exception):
            return WordList.wordlist[lang].index(word)
        raise ValueError('invalid word')


def mnemonic_from_entropy(entropy: Union[bytes, str, None] = None, lang: str = 'en') -> str:
    if entropy:
        assert type(entropy).__name__ in ['bytes', 'str'], 'unsupported entropy type'
        entropy_bytes = entropy if isinstance(entropy, bytes) else bytes.fromhex(entropy)
    else:
        # random a new entropy
        entropy_bytes = randbits(BIP39_ENTROPY_BIT_LENGTH).to_bytes(BIP39_ENTROPY_BIT_LENGTH // 8, 'big')
    entropy_bits: str = bytes_to_bits(entropy_bytes)
    assert len(entropy_bits) in BIP39_ENTROPY_BIT_LENGTH_LIST, 'invalid entropy bit length'
    checksum_bits: str = bytes_to_bits(sha256(entropy_bytes))[:len(entropy_bits) // 32]

    bits: str = entropy_bits + checksum_bits
    indexes_bits: List[str] = [bits[i:i + 11] for i in range(0, len(bits), 11)]
    return ' '.join([WordList.get_word(bits_to_bytes(index_bits), lang) for index_bits in indexes_bits])


def validate_mnemonic(mnemonic: str, lang: str = 'en'):
    indexes: List[int] = [WordList.index_word(word, lang) for word in mnemonic.split(' ')]
    bits: str = ''.join([bin(index)[2:].zfill(11) for index in indexes])
    entropy_bit_length: int = len(bits) * 32 // 33
    assert entropy_bit_length in BIP39_ENTROPY_BIT_LENGTH_LIST, 'invalid mnemonic, bad entropy bit length'
    entropy_bits: str = bits[:entropy_bit_length]
    checksum_bits: str = bytes_to_bits(sha256(bits_to_bytes(entropy_bits)))[:entropy_bit_length // 32]
    assert checksum_bits == bits[entropy_bit_length:], 'invalid mnemonic, checksum mismatch'


def seed_from_mnemonic(mnemonic: str, lang: str = 'en', passphrase: str = '', prefix: str = 'mnemonic') -> bytes:
    validate_mnemonic(mnemonic, lang)
    hash_name = 'sha512'
    password = mnemonic.encode()
    salt = (prefix + passphrase).encode()
    iterations = 2048
    dklen = 64
    return pbkdf2_hmac(hash_name, password, salt, iterations, dklen)



import unittest
from unittest.mock import AsyncMock, MagicMock

from bsv.broadcaster import BroadcastResponse, BroadcastFailure
from bsv.broadcasters.arc import ARC, ARCConfig
from bsv.http_client import HttpClient, HttpResponse, SyncHttpClient
from bsv.transaction import Transaction


class TestARCBroadcast(unittest.IsolatedAsyncioTestCase):

    def setUp(self):
        self.URL = "https://api.taal.com/arc"
        self.api_key = "apikey_85678993923y454i4jhd803wsd02"
        self.tx = Transaction(tx_data="Hello sCrypt")

        # Mocking the Transaction methods
        self.tx.hex = MagicMock(return_value="hexFormat")

    async def test_broadcast_success(self):
        mock_response = HttpResponse(
            ok=True,
            status_code=200,
            json_data={
                "data": {
                    "txid": "8e60c4143879918ed03b8fc67b5ac33b8187daa3b46022ee2a9e1eb67e2e46ec",
                    "txStatus": "success",
                    "extraInfo": "extra",
                }
            },
        )
        mock_http_client = AsyncMock(HttpClient)
        mock_http_client.fetch = AsyncMock(return_value=mock_response)

        arc_config = ARCConfig(api_key=self.api_key, http_client=mock_http_client)
        arc = ARC(self.URL, arc_config)
        result = await arc.broadcast(self.tx)

        self.assertIsInstance(result, BroadcastResponse)
        self.assertEqual(
            result.txid,
            "8e60c4143879918ed03b8fc67b5ac33b8187daa3b46022ee2a9e1eb67e2e46ec",
        )
        self.assertEqual(result.message, "success extra")

    async def test_broadcast_failure(self):
        mock_response = HttpResponse(
            ok=False,
            status_code=400,
            json_data={
                "data": {"status": "ERR_BAD_REQUEST", "detail": "Invalid transaction"}
            },
        )
        mock_http_client = AsyncMock(HttpClient)
        mock_http_client.fetch = AsyncMock(return_value=mock_response)

        arc_config = ARCConfig(api_key=self.api_key, http_client=mock_http_client)
        arc = ARC(self.URL, arc_config)
        result = await arc.broadcast(self.tx)

        self.assertIsInstance(result, BroadcastFailure)
        self.assertEqual(result.code, "400")
        self.assertEqual(result.description, "Invalid transaction")

    async def test_broadcast_exception(self):
        mock_http_client = AsyncMock(HttpClient)
        mock_http_client.fetch = AsyncMock(side_effect=Exception("Internal Error"))

        arc_config = ARCConfig(api_key=self.api_key, http_client=mock_http_client)
        arc = ARC(self.URL, arc_config)
        result = await arc.broadcast(self.tx)

        self.assertIsInstance(result, BroadcastFailure)
        self.assertEqual(result.code, "500")
        self.assertEqual(result.description, "Internal Error")

    def test_sync_broadcast_success(self):
        mock_response = HttpResponse(
            ok=True,
            status_code=200,
            json_data={
                "data": {
                    "txid": "8e60c4143879918ed03b8fc67b5ac33b8187daa3b46022ee2a9e1eb67e2e46ec",
                    "txStatus": "success",
                    "extraInfo": "extra",
                }
            },
        )
        mock_sync_http_client = MagicMock(SyncHttpClient)
        mock_sync_http_client.post = MagicMock(return_value=mock_response)  # fetch → post

        arc_config = ARCConfig(api_key=self.api_key, sync_http_client=mock_sync_http_client)
        arc = ARC(self.URL, arc_config)
        result = arc.sync_broadcast(self.tx)

        self.assertIsInstance(result, BroadcastResponse)
        self.assertEqual(
            result.txid,
            "8e60c4143879918ed03b8fc67b5ac33b8187daa3b46022ee2a9e1eb67e2e46ec",
        )
        self.assertEqual(result.message, "success extra")

    def test_sync_broadcast_failure(self):
        mock_response = HttpResponse(
            ok=False,
            status_code=400,
            json_data={
                "data": {"status": "ERR_BAD_REQUEST", "detail": "Invalid transaction"}
            },
        )
        mock_sync_http_client = MagicMock(SyncHttpClient)
        mock_sync_http_client.post = MagicMock(return_value=mock_response)  # fetch → post

        arc_config = ARCConfig(api_key=self.api_key, sync_http_client=mock_sync_http_client)
        arc = ARC(self.URL, arc_config)
        result = arc.sync_broadcast(self.tx)

        self.assertIsInstance(result, BroadcastFailure)
        self.assertEqual(result.code, "400")
        self.assertEqual(result.description, "Invalid transaction")

    def test_sync_broadcast_timeout_error(self):
        """408 time out error test"""
        mock_response = HttpResponse(
            ok=False,
            status_code=408,
            json_data={"data": {"status": "ERR_TIMEOUT", "detail": "Request timed out"}}
        )
        mock_sync_http_client = MagicMock(SyncHttpClient)
        mock_sync_http_client.post = MagicMock(return_value=mock_response)

        arc_config = ARCConfig(api_key=self.api_key, sync_http_client=mock_sync_http_client)
        arc = ARC(self.URL, arc_config)
        result = arc.sync_broadcast(self.tx, timeout=5)

        self.assertIsInstance(result, BroadcastFailure)
        self.assertEqual(result.status, "failure")
        self.assertEqual(result.code, "408")
        self.assertEqual(result.description, "Transaction broadcast timed out after 5 seconds")

    def test_sync_broadcast_connection_error(self):
        """503 error test"""
        mock_response = HttpResponse(
            ok=False,
            status_code=503,
            json_data={"data": {"status": "ERR_CONNECTION", "detail": "Service unavailable"}}
        )
        mock_sync_http_client = MagicMock(SyncHttpClient)
        mock_sync_http_client.post = MagicMock(return_value=mock_response)

        arc_config = ARCConfig(api_key=self.api_key, sync_http_client=mock_sync_http_client)
        arc = ARC(self.URL, arc_config)
        result = arc.sync_broadcast(self.tx)

        self.assertIsInstance(result, BroadcastFailure)
        self.assertEqual(result.status, "failure")
        self.assertEqual(result.code, "503")
        self.assertEqual(result.description, "Failed to connect to ARC service")

    def test_sync_broadcast_exception(self):
        mock_sync_http_client = MagicMock(SyncHttpClient)
        mock_sync_http_client.post = MagicMock(side_effect=Exception("Internal Error"))

        arc_config = ARCConfig(api_key=self.api_key, sync_http_client=mock_sync_http_client)
        arc = ARC(self.URL, arc_config)
        result = arc.sync_broadcast(self.tx)

        self.assertIsInstance(result, BroadcastFailure)
        self.assertEqual(result.code, "500")
        self.assertEqual(result.description, "Internal Error")

    def test_check_transaction_status_success(self):
        mock_response = HttpResponse(
            ok=True,
            status_code=200,
            json_data={
                "data": {  # dataキーを追加
                    "txid": "8e60c4143879918ed03b8fc67b5ac33b8187daa3b46022ee2a9e1eb67e2e46ec",
                    "txStatus": "MINED",
                    "blockHash": "000000000000000001234567890abcdef",
                    "blockHeight": 800000
                }
            },
        )
        mock_sync_http_client = MagicMock(SyncHttpClient)
        mock_sync_http_client.get = MagicMock(return_value=mock_response)  # fetch → get

        arc_config = ARCConfig(api_key=self.api_key, sync_http_client=mock_sync_http_client)
        arc = ARC(self.URL, arc_config)
        result = arc.check_transaction_status("8e60c4143879918ed03b8fc67b5ac33b8187daa3b46022ee2a9e1eb67e2e46ec")

        self.assertEqual(result["txid"], "8e60c4143879918ed03b8fc67b5ac33b8187daa3b46022ee2a9e1eb67e2e46ec")
        self.assertEqual(result["txStatus"], "MINED")
        self.assertEqual(result["blockHeight"], 800000)

    def test_categorize_transaction_status_mined(self):
        response = {
            "txStatus": "MINED",
            "blockHeight": 800000
        }
        result = ARC.categorize_transaction_status(response)

        self.assertEqual(result["status_category"], "mined")
        self.assertEqual(result["tx_status"], "MINED")

    def test_categorize_transaction_status_progressing(self):
        response = {
            "txStatus": "QUEUED"
        }
        result = ARC.categorize_transaction_status(response)

        self.assertEqual(result["status_category"], "progressing")
        self.assertEqual(result["tx_status"], "QUEUED")

    def test_categorize_transaction_status_warning(self):
        response = {
            "txStatus": "SEEN_ON_NETWORK",
            "competingTxs": ["some_competing_tx"]
        }
        result = ARC.categorize_transaction_status(response)

        self.assertEqual(result["status_category"], "warning")
        self.assertEqual(result["tx_status"], "SEEN_ON_NETWORK")

    def test_categorize_transaction_status_0confirmation(self):
        response = {
            "txStatus": "SEEN_ON_NETWORK"
        }
        result = ARC.categorize_transaction_status(response)

        self.assertEqual(result["status_category"], "0confirmation")
        self.assertEqual(result["tx_status"], "SEEN_ON_NETWORK")


if __name__ == "__main__":
    unittest.main()


from collections import namedtuple
from typing import Optional

from coincurve import PublicKey as CcPublicKey

from .constants import NUMBER_BYTE_LENGTH

Point = namedtuple('Point', 'x y')

EllipticCurve = namedtuple('EllipticCurve', 'name p a b g n h')
curve = EllipticCurve(
    name='Secp256k1',
    p=0xfffffffffffffffffffffffffffffffffffffffffffffffffffffffefffffc2f,
    a=0,
    b=7,
    g=Point(
        0x79be667ef9dcbbac55a06295ce870b07029bfcdb2dce28d959f2815b16f81798,
        0x483ada7726a3c4655da4fbfc0e1108a8fd17b448a68554199c47d08ffb10d4b8
    ),
    n=0xfffffffffffffffffffffffffffffffebaaedce6af48a03bbfd25e8cd0364141,
    h=1,
)


def on_curve(point: Optional[Point]) -> bool:
    """
    :returns: True if the given point lies on the elliptic curve
    """
    if point is None:
        # None represents the point at infinity.
        return True
    x, y = point
    return (y * y - x * x * x - curve.a * x - curve.b) % curve.p == 0


def curve_negative(point: Optional[Point]) -> Optional[Point]:
    """
    :returns: -point
    """
    assert on_curve(point)
    if point is None:
        # -0 = 0
        return None
    x, y = point
    r = Point(x, -y % curve.p)
    assert on_curve(r)
    return r


def curve_add(p: Optional[Point], q: Optional[Point]) -> Optional[Point]:
    """
    :returns: the result of p + q according to the group law
    """
    assert on_curve(p)
    assert on_curve(q)
    if p is None:
        # 0 + q = q
        return q
    if q is None:
        # p + 0 = p
        return p
    if p == curve_negative(q):
        # p == -q
        return None
    # p != -q
    r = Point(*CcPublicKey.from_point(*p).combine([CcPublicKey.from_point(*q)]).point())
    assert on_curve(r)
    return r


def curve_multiply(scalar: int, point: Optional[Point]) -> Optional[Point]:
    """
    multiply the given point by a scalar
    """
    assert on_curve(point)
    if scalar % curve.n == 0 or point is None:
        return None
    if scalar < 0:
        # k * point = -k * (-point)
        return curve_multiply(-scalar, curve_negative(point))
    r = Point(*CcPublicKey.from_point(*point).multiply((scalar % curve.n).to_bytes(NUMBER_BYTE_LENGTH, 'big')).point())
    assert on_curve(r)
    return r


def curve_get_y(x: int, even: bool) -> int:
    """
    point (x, y) lies on the curve, calculate y from the given x and the parity of y
    """
    y_square = (x * x * x + curve.a * x + curve.b) % curve.p
    y = pow(y_square, (curve.p + 1) // 4, curve.p)
    return y if (y + (0 if even else 1)) % 2 == 0 else -y % curve.p
